import numpy as np

import xarray as xr
from arpes.endstations import (FITSEndstation, HemisphericalEndstation,
                               SynchrotronEndstation)

__all__ = ('MAESTROARPESEndstation',)


class MAESTROARPESEndstation(SynchrotronEndstation, HemisphericalEndstation, FITSEndstation):
    """
    The MERLIN ARPES Endstation at the Advanced Light Source
    """

    PRINCIPAL_NAME = 'ALS-BL702'
    ALIASES = ['BL7', 'BL7.0.2', 'ALS-BL7.0.2',]

    RENAME_KEYS = {
        'LMOTOR0': 'x',
        'LMOTOR1': 'y',
        'LMOTOR2': 'z',
        'Scan X': 'scan_x',
        'Scan Y': 'scan_y',
        'Scan Z': 'scan_z',
        'LMOTOR3': 'theta',
        'LMOTOR4': 'beta',
        'LMOTOR5': 'chi',
        'LMOTOR6': 'alpha',
        'LMOTOR9': 'psi',
        'mono_eV': 'hv',
        'SF_HV': 'hv',
        'SS_HV': 'hv',
        'Slit Defl': 'psi',

        # probably need something like an attribute list for extraction
        'SFRGN0': 'fixed_region_name',
        'SFE_0': 'daq_center_energy',
        'SFLNM0': 'lens_mode_name',
        'SFPE_0': 'pass_energy',
        'UNDHARM': 'undulator_harmonic',
        'RINGCURR': 'beam_current',
        'SFFR_0': 'frames_per_slice',

        'SFBA_0': 'phi_prebinning',
        'SFBE0': 'eV_prebinning',
    }

    RENAME_COORDS = {
        'X': 'x',
        'Y': 'y',
        'Z': 'z',
    }

    ATTR_TRANSFORMS = {
        'START_T': lambda l: {'time': ' '.join(l.split(' ')[1:]).lower(),
                              'date': l.split(' ')[0]},
        'SF_SLITN': lambda l: {'slit_number': int(l.split(' ')[0]),
                               'slit_shape': l.split(' ')[-1].lower(),
                               'slit_width': float(l.split(' ')[2])},
    }

    MERGE_ATTRS = {
        'mcp_voltage': None,
        'repetition_rate': 5e8,
        'undulator_type': 'elliptically_polarized_undulator',
        'undulator_gap': None,
        'undulator_z': None,
        'undulator_polarization': None,
    }

    def load(self, scan_desc: dict = None, **kwargs):
        # in the future, can use a regex in order to handle the case where we postfix coordinates
        # for multiple spectra

        scan = super().load(scan_desc, **kwargs)

        coord_names = scan.coords.keys()
        will_rename = {}
        for coord_name in coord_names:
            if coord_name in self.RENAME_KEYS:
                will_rename[coord_name] = self.RENAME_KEYS.get(coord_name)

        renamed = scan.rename(will_rename)

        if 'scan_x' in renamed.coords:
            for d in renamed.data_vars:
                if 'spectrum' in d:
                    renamed[d].values = np.flip(renamed[d].values, axis=renamed[d].dims.index('scan_x'))

        return renamed

    def postprocess_final(self, data: xr.Dataset, scan_desc: dict=None):
        chamber = 'uARPES' if data.attrs['HOST2'] == 'uARPES.als.lbl.gov' else 'nARPES'
        if chamber == 'uARPES':
            analyzer_info = {
                'analyzer': 'R4000',
                'analyzer_name': 'Scienta R4000',
                'parallel_deflectors': False,
                'perpendicular_deflectors': True,
                'analyzer_radius': None,
                'analyzer_type': 'hemispherical',
            }
        else:
            analyzer_info = {
                'analyzer': 'DA-30',
                'analyzer_name': 'Scienta DA-30',
                'parallel_deflectors': False,
                'perpendicular_deflectors': False,
                'analyzer_radius': None,
                'analyzer_type': 'hemispherical',
            }

        ls = [data] + data.S.spectra
        for l in ls:
            l.attrs.update(analyzer_info)

            if 'GRATING' in l.attrs:
                l.attrs['grating_lines_per_mm'] = {
                    'G201b': 600,
                }.get(l.attrs['GRATING'])

        data = super().postprocess_final(data, scan_desc)

        return data



import re
from pathlib import Path

import numpy as np

import typing
import xarray as xr
from arpes.endstations import (HemisphericalEndstation, SESEndstation,
                               SynchrotronEndstation)

__all__ = ('BL403ARPESEndstation',)


class BL403ARPESEndstation(SynchrotronEndstation, HemisphericalEndstation, SESEndstation):
    """
    The MERLIN ARPES Endstation at the Advanced Light Source
    """

    PRINCIPAL_NAME = 'ALS-BL403'
    ALIASES = ['BL403', 'BL4', 'BL4.0.3', 'ALS-BL403', 'ALS-BL4', ]

    RENAME_KEYS = {
        'Polar Compens': 'theta',  # these are caps-ed because they are dimensions in some cases!
        'BL Energy': 'hv',
        'tilt': 'beta', 'polar': 'theta', 'azimuth': 'chi',
        'temperature_sensor_a': 'temperature_cryotip',
        'temperature_sensor_b': 'temperature',
        'cryostat_temp_a': 'temp_cryotip',
        'cryostat_temp_b': 'temp',
        'bl_energy': 'hv',
        'polar_compens': 'theta',
        'K2200 V': 'volts',
        'pwr_supply_v': 'volts',

        'mcp': 'mcp_voltage',
        'slit_plate': 'slit_number',
        'user': 'experimenter',
        'sample': 'sample_name',
        'mesh_current': 'photon_flux',
        'ring_energy': 'beam_energy',
        'epu_pol': 'undulator_polarization',
        'epu_gap': 'undulator_gap',
        'epu_z': 'undulator_z',
        'center_energy': 'daq_center_energy',

        'low_energy': 'sweep_low_energy',
        'high_energy': 'sweep_high_energy',
        'energy_step': 'sweep_step',
        'number_of_sweeps': 'n_sweeps',
    }

    MERGE_ATTRS = {
        'analyzer': 'R8000',
        'analyzer_name': 'Scienta R8000',
        'parallel_deflectors': False,
        'perpendicular_deflectors': False,
        'analyzer_radius': None,
        'analyzer_type': 'hemispherical',
        'repetition_rate': 5e8,
        'undulator_harmonic': 2,  # TODO
        'undulator_type': 'elliptically_polarized_undulator',
    }

    ATTR_TRANSFORMS = {
        'acquisition_mode': lambda l: l.lower(),
        'lens_mode': lambda l: {
            'lens_mode': None,
            'lens_mode_name': l,
        },
        'undulator_polarization': int,
        'region_name': lambda l: {
            'daq_region_name': l,
            'daq_region': l,
        }
    }

    def concatenate_frames(self, frames=typing.List[xr.Dataset], scan_desc: dict = None):
        if len(frames) < 2:
            return super().concatenate_frames(frames)

        # determine which axis to stitch them together along, and then do this
        original_filename = scan_desc.get('file', scan_desc.get('path'))
        assert original_filename is not None

        internal_match = re.match(r'([a-zA-Z0-9\w+_]+)_[S][0-9][0-9][0-9]\.pxt', Path(original_filename).name)
        if internal_match is not None:
            if internal_match.groups():
                motors_path = str(
                    Path(original_filename).parent / '{}_Motor_Pos.txt'.format(internal_match.groups()[0]))
                try:
                    with open(motors_path, 'r') as f:
                        lines = f.readlines()

                    axis_name = lines[0].strip()
                    axis_name = self.RENAME_KEYS.get(axis_name, axis_name)
                    values = [float(l.strip()) for l in lines[1:len(frames) + 1]]

                    for v, f in zip(values, frames):
                        f.coords[axis_name] = v

                    frames.sort(key=lambda x: x.coords[axis_name])

                    for frame in frames:
                        # promote x, y, z to coords so they get concatted
                        for l in [frame] + frame.S.spectra:
                            for c in ['x', 'y', 'z']:
                                if c not in l.coords:
                                    l.coords[c] = l.attrs[c]

                    return xr.concat(frames, axis_name, coords='different')
                except Exception:
                    pass
        else:
            internal_match = re.match(r'([a-zA-Z0-9\w+_]+)_[R][0-9][0-9][0-9]\.pxt', Path(original_filename).name)
            if internal_match.groups():
                return xr.merge(frames)

        return super().concatenate_frames(frames)

    def load_single_frame(self, frame_path: str = None, scan_desc: dict = None, **kwargs):
        import copy
        import os
        from arpes.repair import negate_energy
        from arpes.load_pxt import read_single_pxt, find_ses_files_associated

        _, ext = os.path.splitext(frame_path)
        if 'nc' in ext:
            # was converted to hdf5/NetCDF format with Conrad's Igor scripts
            scan_desc = copy.deepcopy(scan_desc)
            scan_desc['path'] = frame_path
            return self.load_SES_nc(scan_desc=scan_desc, **kwargs)

        original_data_loc = scan_desc.get('path', scan_desc.get('file'))

        p = Path(original_data_loc)

        # find files with same name stem, indexed in format R###
        regions = find_ses_files_associated(p, separator='R')

        if len(regions) == 1:
            pxt_data = negate_energy(read_single_pxt(frame_path))
            return xr.Dataset({'spectrum': pxt_data}, attrs=pxt_data.attrs)
        else:
            # need to merge several different detector 'regions' in the same scan
            region_files = [self.load_single_region(region_path) for region_path in regions]

            # can they share their energy axes?
            all_same_energy = True
            for reg in region_files[1:]:
                dim = 'eV' + reg.attrs['Rnum']
                all_same_energy = all_same_energy and np.array_equal(region_files[0].coords['eV000'], reg.coords[dim])

            if all_same_energy:
                for i, reg in enumerate(region_files):
                    dim = 'eV' + reg.attrs['Rnum']
                    region_files[i] = reg.rename({dim: 'eV'})
            else:
                pass

            return self.concatenate_frames(region_files, scan_desc=scan_desc)

    def load_single_region(self, region_path: str = None, scan_desc: dict = None, **kwargs):
        import os
        from arpes.repair import negate_energy
        from arpes.load_pxt import read_single_pxt

        name, ext = os.path.splitext(region_path)
        num = name[-3:]

        pxt_data = negate_energy(read_single_pxt(region_path))
        pxt_data = pxt_data.rename({'eV': 'eV' + num})
        pxt_data.attrs['Rnum'] = num
        pxt_data.attrs['alpha'] = np.pi / 2
        return xr.Dataset({'spectrum' + num: pxt_data},
                          attrs=pxt_data.attrs)  # separate spectra for possibly unrelated data

    def postprocess_final(self, data: xr.Dataset, scan_desc: dict = None):
        ls = [data] + data.S.spectra

        for l in ls:
            if 'slit_number' in l.attrs:
                slit_lookup = {
                    1: ('straight', 0.1),
                    7: ('curved', 0.5),
                }
                shape, width = slit_lookup.get(l.attrs['slit_number'], (None, None))
                l.attrs['slit_shape'] = shape
                l.attrs['slit_width'] = width

            if 'undulator_polarization' in l.attrs:
                phase_angle_lookup = {
                    0: (0, 0),  # LH
                    2: (np.pi / 2, 0)  # LV
                }
                polarization_theta, polarization_alpha = phase_angle_lookup[int(l.attrs['undulator_polarization'])]
                l.attrs['probe_polarization_theta'] = polarization_theta
                l.attrs['probe_polarization_alpha'] = polarization_alpha

        deg_to_rad_coords = {'theta', 'phi', 'beta', 'chi', 'psi'}
        deg_to_rad_attrs = {'theta', 'beta', 'chi', 'psi', 'alpha'}

        for c in deg_to_rad_coords:
            if c in data.dims:
                data.coords[c] = data.coords[c] * np.pi / 180

        for angle_attr in deg_to_rad_attrs:
            for l in ls:
                if angle_attr in l.attrs:
                    l.attrs[angle_attr] = float(l.attrs[angle_attr]) * np.pi / 180

        data.attrs['alpha'] = np.pi / 2
        data.attrs['psi'] = 0
        for s in data.S.spectra:
            s.attrs['alpha'] = np.pi / 2
            s.attrs['psi'] = 0

        data = super().postprocess_final(data, scan_desc)

        return data



import numpy as np

from arpes.plotting.utils import name_for_dim, unit_for_dim
from arpes.utilities.conversion.forward import convert_coordinates_to_kspace_forward

__all__ = ('annotate_cuts', 'annotate_point', 'annotate_experimental_conditions',)

def annotate_experimental_conditions(ax, data, desc, show=False, orientation='top', **kwargs):
    """
    Renders information about the experimental conditions onto a set of axes,
    also adjust the axes limits and hides the axes.

    data should be the dataset described, and desc should be one of

    'temp',
    'photon',
    'photon polarization',
    'polarization',
    or a number to act as a spacer in units of the axis coordinates

    or a list of such items

    :param ax:
    :param data:
    :param desc:
    :return:
    """

    if isinstance(desc, (str, int, float,)):
        desc = [desc]

    ax.grid(False)
    ax.set_ylim([0, 100])
    ax.set_xlim([0, 100])
    if not show:
        ax.set_axis_off()
        ax.patch.set_alpha(0)

    delta = -1
    current = 100
    if orientation == 'bottom':
        delta = 1
        current = 0

    fontsize = kwargs.pop('fontsize', 16)
    delta = fontsize * delta

    conditions = data.S.experimental_conditions

    def render_polarization(c):
        pol = c['polarization']
        if pol in ['lc', 'rc']:
            return '\\textbf{' + pol.upper() + '}'

        symbol_pol = {
            's': '',
            'p': '',
            's-p': '',
            'p-s': '',
        }

        prefix = ''
        if pol in ['s-p', 'p-s']:
            prefix = '\\textbf{Linear Dichroism, }'

        symbol = symbol_pol[pol]
        if symbol:
            return prefix + '$' + symbol + '$/\\textbf{' + pol + '}'

        return prefix + '\\textbf{' + pol + '}'

    def render_photon(c):
        return '\\textbf{' + str(c['hv']) + ' eV'

    renderers = {
        'temp': lambda c: '\\textbf{T = ' + '{:.3g}'.format(c['temp']) + ' K}',
        'photon': render_photon,
        'photon polarization': lambda c: render_photon(c) + ', ' + render_polarization(c),
        'polarization': render_polarization,
    }

    for item in desc:
        if isinstance(item, (float, int)):
            current += item + delta
            continue

        item = item.replace('_', ' ').lower()

        ax.text(0, current, renderers[item](conditions), fontsize=fontsize, **kwargs)
        current += delta


def annotate_cuts(ax, data, plotted_axes, include_text_labels=False, **kwargs):
    """
    Example: annotate_cuts(ax, conv, ['kz', 'ky'], hv=80)

    :param ax:
    :param data:
    :param plotted_axes:
    :param include_text_labels:
    :param kwargs:
    :return:
    """
    converted_coordinates = convert_coordinates_to_kspace_forward(data)
    assert len(plotted_axes) == 2

    for k, v in kwargs.items():
        if not isinstance(v, (tuple, list, np.ndarray,)):
            v = [v]

        selected = converted_coordinates.sel(**dict([[k, v]]), method='nearest')

        for coords_dict, obj in selected.T.iterate_axis(k):
            css = [obj[d].values for d in plotted_axes]
            ax.plot(*css, color='red', ls='--', linewidth=1, dashes=(5, 5))

            if include_text_labels:
                idx = np.argmin(css[1])

                ax.text(css[0][idx] + 0.05, css[1][idx], '{} = {} {}'.format(name_for_dim(k), coords_dict[k].item(), unit_for_dim(k)),
                        color='red', size='medium')


def annotate_point(ax, location, label, delta=None, **kwargs):
    label = {
        'G': '$\\Gamma$',
        'X': r'\textbf{X}',
        'Y': r'\textbf{Y}',
        'K': r'\textbf{K}',
        'M': r'\textbf{M}',
    }.get(label, label)

    if delta is None:
        delta = (-0.05, 0.05,)

    c = kwargs.pop('color', 'red')

    if len(delta) == 2:
        dx, dy = tuple(delta)
        x, y = tuple(location)
        ax.plot([x], [y], 'o', c=c)
        ax.text(x + dx, y + dy, label, color=c, **kwargs)
    else:
        dx, dy, dz = tuple(delta)
        x, y, z = tuple(location)
        ax.plot([x], [y], [z], 'o', c=c)
        ax.text(x + dx, y + dy, z + dz, label, color=c, **kwargs)



import functools
import warnings

import numpy as np

import xarray as xr
from arpes.provenance import provenance, update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('curvature', 'dn_along_axis', 'd2_along_axis', 'd1_along_axis', 'minimum_gradient', 'vector_diff')


def vector_diff(arr, delta, n=1):
    """
    Computes finite differences along the vector delta, given as a tuple

    Using delta = (0, 1) is equivalent to np.diff(..., axis=1), while
    using delta = (1, 0) is equivalent to np.diff(..., axis=0).
    :param arr: np.ndarray
    :param delta: iterable containing vector to take difference along
    :return:
    """

    if n == 0:
        return arr
    if n < 0:
        raise ValueError(
            'Order must be non-negative but got ' + repr(n))

    nd = arr.ndim
    slice1 = [slice(None)] * nd
    slice2 = [slice(None)] * nd

    for dim, delta_val in enumerate(delta):
        if delta_val != 0:
            if delta_val < 0:
                slice2[dim] = slice(-delta_val, None)
                slice1[dim] = slice(None, delta_val)
            else:
                slice1[dim] = slice(delta_val, None)
                slice2[dim] = slice(None, -delta_val)

    if n > 1:
        return vector_diff(arr[slice1] - arr[slice2], delta, n - 1)

    return arr[slice1] - arr[slice2]


@update_provenance('Minimum Gradient')
def minimum_gradient(data: DataType,delta=1):
    arr = normalize_to_spectrum(data)
    new = arr / gradient_modulus(arr,delta=delta)
    new.values[np.isnan(new.values)] = 0
    return new


@update_provenance('Gradient Modulus')
def gradient_modulus(data: DataType, delta=1):
    spectrum = normalize_to_spectrum(data)
    values = spectrum.values
    gradient_vector = np.zeros(shape=(8,) + values.shape)

    gradient_vector[0, :-delta, :] = vector_diff(values, (delta, 0,))
    gradient_vector[1, :, :-delta] = vector_diff(values, (0, delta,))
    gradient_vector[2, delta:, :] = vector_diff(values, (-delta, 0,))
    gradient_vector[3, :, delta:] = vector_diff(values, (0, -delta,))
    gradient_vector[4, :-delta, :-delta] = vector_diff(values, (delta, delta,))
    gradient_vector[5, :-delta, delta:] = vector_diff(values, (delta, -delta,))
    gradient_vector[6, delta:, :-delta] = vector_diff(values, (-delta, delta,))
    gradient_vector[7, delta:, delta:] = vector_diff(values, (-delta, -delta,))

    data_copy = spectrum.copy(deep=True)
    data_copy.values = np.linalg.norm(gradient_vector, axis=0)
    return data_copy


def curvature(arr: xr.DataArray, directions=None, alpha=1, beta=None):
    """
    Defined via
        C(x,y) = ([C_0 + (df/dx)^2]d^2f/dy^2 - 2 * df/dx df/dy d^2f/dxdy + [C_0 + (df/dy)^2]d^2f/dx^2) /
                 (C_0 (df/dx)^2 + (df/dy)^2)^(3/2)

    of in the case of inequivalent dimensions x and y

        C(x,y) = ([1 + C_x(df/dx)^2]C_y * d^2f/dy^2 -
                  2 * C_x * C_y * df/dx df/dy d^2f/dxdy +
                  [1 + C_y * (df/dy)^2] * C_x * d^2f/dx^2) /
                 (1 + C_x (df/dx)^2 + C_y (df/dy)^2)^(3/2)

        where
        C_x = C_y * (xi / eta)^2
        and where (xi / eta) = dx / dy

        The value of C_y can reasonably be taken to have the value |df/dx|_max^2 + |df/dy|_max^2
        C_y = (dy / dx) * (|df/dx|_max^2 + |df/dy|_max^2) * \alpha

        for some dimensionless parameter alpha
    :param arr:
    :param alpha: regulation parameter, chosen semi-universally, but with no particular justification
    :return:
    """
    if beta is not None:
        alpha = np.power(10., beta)

    if directions is None:
        directions = arr.dims[:2]

    axis_indices = tuple(arr.dims.index(d) for d in directions)
    dx, dy = tuple(float(arr.coords[d][1] - arr.coords[d][0]) for d in directions)
    dfx, dfy = np.gradient(arr.values, dx, dy, axis=axis_indices)
    np.nan_to_num(dfx, copy=False)
    np.nan_to_num(dfy, copy=False)

    mdfdx, mdfdy = np.max(np.abs(dfx)), np.max(np.abs(dfy))

    cy = (dy / dx) * (mdfdx ** 2 + mdfdy ** 2) * alpha
    cx = (dx / dy) * (mdfdx ** 2 + mdfdy ** 2) * alpha

    dfx_2, dfy_2 = np.power(dfx, 2), np.power(dfy, 2)
    d2fy = np.gradient(dfy, dy, axis=axis_indices[1])
    d2fx = np.gradient(dfx, dx, axis=axis_indices[0])
    d2fxy = np.gradient(dfx, dy, axis=axis_indices[1])

    denom = np.power((1 + cx * dfx_2 + cy * dfy_2), 1.5)
    numerator = (1 + cx * dfx_2) * cy * d2fy - 2 * cx * cy * dfx * dfy * d2fxy + \
                (1 + cy * dfy_2) * cx * d2fx

    curv = xr.DataArray(
        numerator / denom,
        arr.coords,
        arr.dims,
        attrs=arr.attrs
    )

    if 'id' in curv.attrs:
        del curv.attrs['id']
        provenance(curv, arr, {
            'what': 'Curvature',
            'by': 'curvature',
            'directions': directions,
            'alpha': alpha,
        })
    return curv


def dn_along_axis(arr: xr.DataArray, axis=None, smooth_fn=None, order=2):
    """
    Like curvature, performs a second derivative. You can pass a function to use for smoothing through
    the parameter smooth_fn, otherwise no smoothing will be performed.

    You can specify the axis to take the derivative along with the axis param, which expects a string.
    If no axis is provided the axis will be chosen from among the available ones according to the preference
    for axes here, the first available being taken:

    ['eV', 'kp', 'kx', 'kz', 'ky', 'phi', 'beta', 'theta]
    :param arr:
    :param axis:
    :param smooth_fn:
    :param order: Specifies how many derivatives to take
    :return:
    """
    axis_order = ['eV', 'kp', 'kx', 'kz', 'ky', 'phi', 'beta', 'theta']
    if axis is None:
        axes = [a for a in axis_order if a in arr.dims]
        if axes:
            axis = axes[0]
        else:
            # have to do something
            axis = arr.dims[0]
            warnings.warn('Choosing axis: {} for the second derivative, no preferred axis found.'.format(axis))

    if smooth_fn is None:
        smooth_fn = lambda x: x

    d_axis = float(arr.coords[axis][1] - arr.coords[axis][0])
    axis_idx = arr.dims.index(axis)

    values = arr.values
    for _ in range(order):
        as_arr = xr.DataArray(values, arr.coords, arr.dims)
        smoothed = smooth_fn(as_arr)
        values = np.gradient(smoothed.values, d_axis, axis=axis_idx)

    dn_arr = xr.DataArray(
        values,
        arr.coords,
        arr.dims,
        attrs=arr.attrs
    )

    if 'id' in dn_arr.attrs:
        del dn_arr.attrs['id']
        provenance(dn_arr, arr, {
            'what': '{}th derivative'.format(order),
            'by': 'dn_along_axis',
            'axis': axis,
            'order': order,
        })

    return dn_arr


d2_along_axis = functools.partial(dn_along_axis, order=2)
d1_along_axis = functools.partial(dn_along_axis, order=1)



import numpy as np
from sklearn.decomposition import PCA

import xarray as xr
from arpes.fits.fit_models import AffineBackgroundModel, LorentzianModel
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.conversion import slice_along_path

__all__ = ('curves_along_pocket', 'edcs_along_pocket', 'radial_edcs_along_pocket',
           'pocket_parameters',)


def pocket_parameters(data: DataType, kf_method=None, sel=None, method_kwargs=None, **kwargs):
    """
    Estimates pocket center, anisotropy, principal vectors, and extent in either
    angle or k-space. Since data can be converted forward it is generally advised to do
    this analysis in angle space before conversion.
    :param data:
    :param kf_method:
    :param sel:
    :param method_kwargs:
    :param kwargs:
    :return:
    """
    slices, _ = curves_along_pocket(data, **kwargs) # slices, angles =

    if kf_method is None:
        kf_method = find_kf_by_mdc

    if sel is None:
        sel = {'eV': slice(-0.03, 0.05)}

    kfs = [kf_method(s if sel is None else s.sel(**sel), **(method_kwargs or {})) for s in slices]

    fs_dims = list(data.dims)
    if 'eV' in fs_dims:
        fs_dims.remove('eV')

    locations = [{d: ss[d].sel(angle=kf, eV=0, method='nearest').item() for d in fs_dims} for kf, ss in
                 zip(kfs, slices)]

    location_vectors = [[coord[d] for d in fs_dims] for coord in locations]
    as_ndarray = np.array(location_vectors)

    pca = PCA(n_components=2)
    pca.fit(as_ndarray)

    return {
        'locations': locations,
        'location_vectors': location_vectors,
        'center': {d: np.mean(np.array([coord[d] for coord in locations])) for d in fs_dims},
        'pca': pca.components_,
    }


@update_provenance('Collect EDCs projected at an angle from pocket')
def radial_edcs_along_pocket(data: DataType, angle, inner_radius=0, outer_radius=5,
                             n_points=None, select_radius=None, **kwargs):
    """
    Produces EDCs distributed radially along a vector from the pocket center. The pocket center
    should be passed through kwargs via `{dim}={value}`. I.e. an appropriate call would be

    radial_edcs_along_pocket(spectrum, np.pi / 4, inner_radius=1, outer_radius=4, phi=0.1, beta=0)

    :param data: ARPES Spectrum
    :param angle: Angle along the FS to cut against
    :param n_points: Number of EDCs, can be automatically inferred
    :param select_radius:
    :param kwargs:
    :return:
    """

    data = normalize_to_spectrum(data)
    fermi_surface_dims = list(data.dims)

    assert 'eV' in fermi_surface_dims
    fermi_surface_dims.remove('eV')

    center_point = {k: v for k, v in kwargs.items() if k in data.dims}
    center_as_vector = np.array([center_point.get(d, 0) for d in fermi_surface_dims])

    if n_points is None:
        stride = data.T.stride(generic_dim_names=False)
        granularity = np.mean(np.array([stride[d] for d in fermi_surface_dims]))
        n_points = int(1. * (outer_radius - inner_radius) / granularity)

    if n_points <= 0:
        n_points = 10

    primitive = np.array([np.cos(angle), np.sin(angle)])
    far = center_as_vector + outer_radius * primitive
    near = center_as_vector + inner_radius * primitive
    vecs = zip(near, far)

    radius_coord = np.linspace(inner_radius, outer_radius, n_points or 10)

    data_vars = {}
    for d, points in dict(zip(fermi_surface_dims, vecs)).items():
        print(d, points)
        data_vars[d] = xr.DataArray(
            np.array(np.linspace(points[0], points[1], n_points)),
            coords={'r': radius_coord}, dims=['r']
        )

    selection_coords = [{k: v[n] for k, v in data_vars.items()} for n in range(n_points)]

    edcs = [data.S.select_around(coord, radius=select_radius, fast=True)
            for coord in selection_coords]

    for r, edc in zip(radius_coord, edcs):
        edc.coords['r'] = r

    data_vars['data'] = xr.concat(edcs, dim='r')

    return xr.Dataset(data_vars, coords=data_vars['data'].coords)


def curves_along_pocket(data: DataType, n_points=None, inner_radius=0, outer_radius=5, shape=None,
                        **kwargs):
    """
    Produces radial slices along a Fermi surface through a pocket. Evenly distributes perpendicular cuts along an
    ellipsoid. The major axes of the ellipsoid can be specified by `shape` but must be axis aligned.

    The inner and outer radius parameters control the endpoints of the resultant slices along the Fermi surface
    :param data:
    :param n_points:
    :param inner_radius:
    :param outer_radius:
    :param shape:
    :param kwargs:
    :return:
    """
    data = normalize_to_spectrum(data)

    fermi_surface_dims = list(data.dims)
    if 'eV' in fermi_surface_dims:
        fermi_surface_dims.remove('eV')

    center_point = {k: v for k, v in kwargs.items() if k in data.dims}

    center_as_vector = np.array([center_point.get(d, 0) for d in fermi_surface_dims])

    if n_points is None:
        # determine N approximately by the granularity
        n_points = np.min(list(len(data.coords[d].values) for d in fermi_surface_dims))

    stride = data.T.stride(generic_dim_names=False)
    resolution = np.min([v for s, v in stride.items() if s in fermi_surface_dims])

    angles = np.linspace(0, 2*np.pi, n_points, endpoint=False)

    def slice_at_angle(theta):
        primitive = np.array([np.cos(theta), np.sin(theta)])
        far = center_as_vector + outer_radius * primitive

        return slice_along_path(data, [dict(zip(fermi_surface_dims, point))
                                       for point in [center_as_vector, far]], resolution=resolution)

    slices = [slice_at_angle(theta) for theta in angles]

    max_ang = slices[0].coords['angle'].max().item()

    slices = [s.sel(angle=slice(max_ang * (1.0 * inner_radius / outer_radius), None)).isel(angle=slice(None, -1))
              for s in slices]

    for ang, s in zip(angles, slices):
        s.coords['theta'] = ang

    # we do not xr.concat because the interpolated angular dim can actually be different on each due
    # to floating point nonsense
    return slices, angles


def find_kf_by_mdc(slice: DataType, offset=0, **kwargs):
    """
    Offset is used to control the radial offset from the pocket for studies where
    you want to go slightly off the Fermi momentum
    :param slice:
    :param offset:
    :param kwargs:
    :return:
    """
    if isinstance(slice, xr.Dataset):
        slice = slice.data

    assert isinstance(slice, xr.DataArray)

    if 'eV' in slice.dims:
        slice = slice.sum('eV')

    lor = LorentzianModel()
    bkg = AffineBackgroundModel(prefix='b_')

    result = (lor + bkg).guess_fit(data=slice, params=kwargs)
    return result.params['center'].value + offset


@update_provenance('Collect EDCs around pocket edge')
def edcs_along_pocket(data: DataType, kf_method=None, select_radius=None,
                      sel=None, method_kwargs=None, **kwargs):
    """
    Collects EDCs around a pocket. This consists first in identifying the momenta
    around the pocket, and then integrating small windows around each of these points.

    :param data:
    :param kf_method:
    :param select_radius:
    :param sel:
    :param method_kwargs:
    :param kwargs:
    :return:
    """
    slices, angles = curves_along_pocket(data, **kwargs)

    if kf_method is None:
        kf_method = find_kf_by_mdc

    if sel is None:
        sel = {'eV': slice(-0.05, 0.05)}

    kfs = [kf_method(s if sel is None else s.sel(**sel), **(method_kwargs or {})) for s in slices]

    fs_dims = list(data.dims)
    if 'eV' in fs_dims:
        fs_dims.remove('eV')

    locations = [{d: ss[d].sel(angle=kf, eV=0, method='nearest').item() for d in fs_dims} for kf, ss in
                 zip(kfs, slices)]

    edcs = [data.S.select_around(l, radius=select_radius, fast=True)
            for l in locations]

    data_vars = {}
    index = np.array(angles)

    for d in fs_dims:
        data_vars[d] = xr.DataArray(
            np.array([l[d] for l in locations]), coords={'theta': index}, dims=['theta'])

    for ang, edc in zip(angles, edcs):
        edc.coords['theta'] = ang

    data_vars['spectrum'] = xr.concat(edcs, dim='theta')

    return xr.Dataset(data_vars, coords=data_vars['spectrum'].coords)



import matplotlib.patches
import matplotlib.path
import matplotlib.pyplot as plt
import numpy as np

import xarray as xr
from arpes.plotting.utils import path_for_holoviews, path_for_plot
from arpes.provenance import save_plot_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('fermi_surface_slices', 'magnify_circular_regions_plot',)

@save_plot_provenance
def fermi_surface_slices(arr: xr.DataArray, n_slices=9, ev_per_slice=0.02, bin=0.01, out=None, **kwargs):
    import holoviews as hv  # pylint: disable=import-error
    slices = []
    for i in range(n_slices):
        high = - ev_per_slice * i
        low = high - bin
        image = hv.Image(arr.sum([d for d in arr.dims if d not in ['theta', 'beta', 'phi', 'eV', 'kp', 'kx', 'ky']]).sel(
            eV=slice(low, high)).sum('eV'), label='%g eV' % high)

        slices.append(image)

    layout = hv.Layout(slices).cols(3)
    if out is not None:
        renderer = hv.renderer('matplotlib').instance(fig='svg', holomap='gif')
        filename = path_for_plot(out)
        renderer.save(layout, path_for_holoviews(filename))
        return filename
    else:
        return layout


@save_plot_provenance
def magnify_circular_regions_plot(data: DataType, magnified_points, mag=10, radius=0.05, cmap='viridis', color=None, edgecolor='red', out=None, ax=None, **kwargs):
    data = normalize_to_spectrum(data)
    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get('figsize', (7, 5,)))

    mesh = data.plot(ax=ax, cmap=cmap)
    clim = list(mesh.get_clim())
    clim[1] = clim[1] / mag

    mask = np.zeros(shape=(len(data.values.ravel()),))
    pts = np.zeros(shape=(len(data.values.ravel()), 2,))
    mask = mask > 0

    raveled = data.T.ravel()
    pts[:,0] = raveled[data.dims[0]]
    pts[:,1] = raveled[data.dims[1]]

    x0, y0 = ax.transAxes.transform((0, 0))  # lower left in pixels
    x1, y1 = ax.transAxes.transform((1, 1))  # upper right in pixes
    dx = x1 - x0
    dy = y1 - y0
    maxd = max(dx, dy)
    xlim, ylim = ax.get_xlim(), ax.get_ylim()

    width = radius * maxd / dx * (xlim[1] - xlim[0])
    height = radius * maxd / dy * (ylim[1] - ylim[0])

    if not isinstance(edgecolor, list):
        edgecolor = [edgecolor for _ in range(len(magnified_points))]

    if not isinstance(color, list):
        color = [color for _ in range(len(magnified_points))]

    pts[:,1] = (pts[:,1]) / (xlim[1] - xlim[0])
    pts[:,0] = (pts[:,0]) / (ylim[1] - ylim[0])
    print(np.min(pts[:, 1]), np.max(pts[:, 1]))
    print(np.min(pts[:, 0]), np.max(pts[:, 0]))

    for c, ec, point in zip(color, edgecolor, magnified_points):
        patch = matplotlib.patches.Ellipse(point, width, height, color=c, edgecolor=ec, fill=False, linewidth=2, zorder=4)
        patchfake = matplotlib.patches.Ellipse(
            [point[1], point[0]],
            radius, radius
        )
        ax.add_patch(patch)
        mask = np.logical_or(mask, patchfake.contains_points(pts))


    data_masked = data.copy(deep=True)
    data_masked.values = np.array(data_masked.values, dtype=np.float32)

    cm = matplotlib.cm.get_cmap(name='viridis')
    cm.set_bad(color=(1, 1, 1, 0))
    data_masked.values[np.swapaxes(np.logical_not(mask.reshape(data.values.shape[::-1])), 0, 1)] = np.nan

    aspect = ax.get_aspect()
    extent = [xlim[0], xlim[1], ylim[0], ylim[1]]
    ax.imshow(data_masked.values, cmap=cm, extent=extent, zorder=3, clim=clim, origin='lower')
    ax.set_aspect(aspect)

    for spine in ['left', 'top', 'right', 'bottom']:
        ax.spines[spine].set_zorder(5)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax



import warnings
from math import factorial

import numpy as np
import scipy.signal

import typing
import xarray as xr
from arpes.provenance import update_provenance
from arpes.typing import DataType

__all__ = ('savitzky_golay',)


@update_provenance('Savitzky Golay Filter')
def savitzky_golay(data: typing.Union[DataType, list, np.ndarray], window_size, order, deriv=0, rate=1, dim=None):
    """
    Implements a Savitzky Golay filter with given window size. You can specify "pass through" dimensions
    which will not be touched with the `dim` argument. This allows for filtering each frame of a map or each equal-energy
    contour in a 3D dataset, for instance.

    :param data: Input data.
    :param window_size: Number of points in the window that the filter uses locally.
    :param order: The polynomial order used in the convolution.
    :param deriv:
    :param rate:
    :param dim:
    :return:
    """
    if isinstance(data, (list, np.ndarray,)):
        return savitzky_golay_array(data, window_size, order, deriv, rate)

    if len(data.dims) == 1:
        transformed_data = savitzky_golay_array(data.values, window_size, order, deriv, rate)
    else:
        # only 1D, 2D, 3D supported for the moment
        assert len(data.dims) <= 3

        if deriv == 0:
            deriv = None

        if len(data.dims) == 3:
            if dim is None:
                dim = data.dims[-1]
            return data.T.map_axes(dim, lambda d, _: savitzky_golay(d, window_size, order, deriv, rate))

        if len(data.dims) == 2:
            if dim is None:
                transformed_data = savitzky_golay_2d(data.values, window_size, order, derivative=deriv)
            else:
                return data.T.map_axes(dim, lambda d, _: savitzky_golay(d, window_size, order, deriv=deriv or 0, rate=rate, dim=None))

    return xr.DataArray(transformed_data, data.coords, data.dims, attrs=data.attrs.copy())


def savitzky_golay_2d(z, window_size, order, derivative=None):
    """
    Implementation from the scipy cookbook before the Savitzky Golay filter
    was supported. This is changed now, so we should ideally migrate to use the
    new scipy implementation.

    :param z:
    :param window_size:
    :param order:
    :param derivative:
    :return:
    """
    # number of terms in the polynomial expression
    n_terms = (order + 1) * (order + 2) / 2.0

    if window_size % 2 == 0 or window_size < 1:
        warnings.warn("window_size size must be a positive odd number")
        window_size = np.max([0, window_size])
        window_size += 1

    if window_size ** 2 < n_terms:
        raise ValueError('order is too high for the window size')

    half_size = window_size // 2

    # exponents of the polynomial.
    # p(x,y) = a0 + a1*x + a2*y + a3*x^2 + a4*y^2 + a5*x*y + ...
    # this line gives a list of two item tuple. Each tuple contains
    # the exponents of the k-th term. First element of tuple is for x
    # second element for y.
    # Ex. exps = [(0,0), (1,0), (0,1), (2,0), (1,1), (0,2), ...]
    exps = [(k - n, n) for k in range(order + 1) for n in range(k + 1)]

    # coordinates of points
    ind = np.arange(-half_size, half_size + 1, dtype=np.float64)
    dx = np.repeat(ind, window_size)
    dy = np.tile(ind, [window_size, 1]).reshape(window_size ** 2, )

    # build matrix of system of equation
    A = np.empty((window_size ** 2, len(exps)))
    for i, exp in enumerate(exps):
        A[:, i] = (dx ** exp[0]) * (dy ** exp[1])

    # pad input array with appropriate values at the four borders
    new_shape = z.shape[0] + 2 * half_size, z.shape[1] + 2 * half_size
    Z = np.zeros((new_shape))
    # top band
    band = z[0, :]
    Z[:half_size, half_size:-half_size] = band - np.abs(np.flipud(z[1:half_size + 1, :]) - band)
    # bottom band
    band = z[-1, :]
    Z[-half_size:, half_size:-half_size] = band + np.abs(np.flipud(z[-half_size - 1:-1, :]) - band)
    # left band
    band = np.tile(z[:, 0].reshape(-1, 1), [1, half_size])
    Z[half_size:-half_size, :half_size] = band - np.abs(np.fliplr(z[:, 1:half_size + 1]) - band)
    # right band
    band = np.tile(z[:, -1].reshape(-1, 1), [1, half_size])
    Z[half_size:-half_size, -half_size:] = band + np.abs(np.fliplr(z[:, -half_size - 1:-1]) - band)
    # central band
    Z[half_size:-half_size, half_size:-half_size] = z

    # top left corner
    band = z[0, 0]
    Z[:half_size, :half_size] = band - np.abs(np.flipud(np.fliplr(z[1:half_size + 1, 1:half_size + 1])) - band)
    # bottom right corner
    band = z[-1, -1]
    Z[-half_size:, -half_size:] = band + np.abs(
        np.flipud(np.fliplr(z[-half_size - 1:-1, -half_size - 1:-1])) - band)

    # top right corner
    band = Z[half_size, -half_size:]
    Z[:half_size, -half_size:] = band - np.abs(np.flipud(Z[half_size + 1:2 * half_size + 1, -half_size:]) - band)
    # bottom left corner
    band = Z[-half_size:, half_size].reshape(-1, 1)
    Z[-half_size:, :half_size] = band - np.abs(np.fliplr(Z[-half_size:, half_size + 1:2 * half_size + 1]) - band)

    # solve system and convolve
    if derivative is None:
        m = np.linalg.pinv(A)[0].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, m, mode='valid')
    elif derivative == 'col':
        c = np.linalg.pinv(A)[1].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -c, mode='valid')
    elif derivative == 'row':
        r = np.linalg.pinv(A)[2].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -r, mode='valid')
    elif derivative == 'both':
        c = np.linalg.pinv(A)[1].reshape((window_size, -1))
        r = np.linalg.pinv(A)[2].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -r, mode='valid'), scipy.signal.fftconvolve(Z, -c, mode='valid')


def savitzky_golay_array(y, window_size, order, deriv=0, rate=1):
    r"""Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.
    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.
    Examples
    --------
    t = np.linspace(-4, 4, 500)
    y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()
    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """

    try:
        window_size = np.abs(np.int(window_size))
        order = np.abs(np.int(order))
    except ValueError:
        raise ValueError("window_size and order have to be of type int")

    if window_size % 2 != 1 or window_size < 1:
        warnings.warn("window_size size must be a positive odd number")
        window_size = np.max([0, window_size])
        window_size += 1

    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")

    order_range = range(order + 1)
    half_window = (window_size - 1) // 2
    # precompute coefficients
    b = np.mat([[k ** i for i in order_range] for k in range(-half_window, half_window + 1)])
    m = np.linalg.pinv(b).A[deriv] * rate ** deriv * factorial(deriv)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs(y[1:half_window + 1][::-1] - y[0])
    lastvals = y[-1] + np.abs(y[-half_window - 1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve(m[::-1], y, mode='valid')



import inspect

import numpy as np

import typing
from arpes.exceptions import AnalysisError
from arpes.plotting.interactive_utils import BokehInteractiveTool, CursorTool
from arpes.typing import DataType
from arpes.utilities import Debounce, normalize_to_spectrum

__all__ = ('DynamicTool', 'dyn',)


class DynamicTool(BokehInteractiveTool, CursorTool):
    """
    Presents a utility to rerun a function with different arguments and see the result of the function
    """

    def __init__(self, analysis_fn, widget_specification, **kwargs):
        super().__init__()

        self.load_settings(**kwargs)
        self.analysis_fn = analysis_fn
        self.widget_specification = widget_specification

        self.app_main_size = self.settings.get('main_width', 600)
        self.app_marginal_size = self.settings.get('marginal_width', 300)

    def tool_handler(self, doc):
        from bokeh import events
        from bokeh.layouts import row, column, widgetbox
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets
        from bokeh.plotting import figure

        if len(self.arr.shape) != 2:
            raise AnalysisError('Cannot use the band tool on non image-like spectra')

        self.data_for_display = self.arr
        x_coords, y_coords = self.data_for_display.coords[self.data_for_display.dims[0]], self.data_for_display.coords[self.data_for_display.dims[1]]

        default_palette = self.default_palette

        self.app_context.update({
            'data': self.arr,
            'data_range': {
                'x': (np.min(x_coords.values), np.max(x_coords.values)),
                'y': (np.min(y_coords.values), np.max(y_coords.values)),
            },
        })

        figures, plots = self.app_context['figures'], self.app_context['plots']

        self.cursor = [np.mean(self.data_range['x']),
                       np.mean(self.data_range['y'])]

        self.color_maps['main'] = LinearColorMapper(
            default_palette, low=np.min(self.data_for_display.values), high=np.max(self.data_for_display.values), nan_color='black')

        main_tools = ["wheel_zoom", "tap", "reset"]
        main_title = '{} Tool: WARNING Unidentified'.format(self.analysis_fn.__name__)

        try:
            main_title = '{} Tool: {}'.format(self.analysis_fn.__name__, self.data_for_display.S.label[:60])
        except:
            pass

        figures['main'] = figure(
            tools=main_tools, plot_width=self.app_main_size, plot_height=self.app_main_size, min_border=10,
            min_border_left=50,
            toolbar_location='left', x_axis_location='below', y_axis_location='right',
            title=main_title, x_range=self.data_range['x'],
            y_range=self.data_range['y'])
        figures['main'].xaxis.axis_label = self.data_for_display.dims[0]
        figures['main'].yaxis.axis_label = self.data_for_display.dims[1]
        figures['main'].toolbar.logo = None
        figures['main'].background_fill_color = "#fafafa"
        plots['main'] = figures['main'].image(
            [self.data_for_display.values.T], x=self.app_context['data_range']['x'][0], y=self.app_context['data_range']['y'][0],
            dw=self.app_context['data_range']['x'][1] - self.app_context['data_range']['x'][0],
            dh=self.app_context['data_range']['y'][1] - self.app_context['data_range']['y'][0],
            color_mapper=self.app_context['color_maps']['main']
        )

        # Create the bottom marginal plot
        bottom_marginal = self.data_for_display.sel(**dict([[self.data_for_display.dims[1], self.cursor[1]]]),
                                                    method='nearest')
        bottom_marginal_original = self.arr.sel(**dict([[self.data_for_display.dims[1], self.cursor[1]]]),
                                                    method='nearest')
        figures['bottom_marginal'] = figure(plot_width=self.app_main_size, plot_height=200,
                                            title=None, x_range=figures['main'].x_range,
                                            y_range=(
                                                np.min(bottom_marginal.values), np.max(bottom_marginal.values)),
                                            x_axis_location='above', toolbar_location=None, tools=[])
        plots['bottom_marginal'] = figures['bottom_marginal'].line(
            x=bottom_marginal.coords[self.data_for_display.dims[0]].values, y=bottom_marginal.values)
        plots['bottom_marginal_original'] = figures['bottom_marginal'].line(
            x=bottom_marginal_original.coords[self.arr.dims[0]].values, y=bottom_marginal_original.values,
            line_color='red')

        # Create the right marginal plot
        right_marginal = self.data_for_display.sel(**dict([[self.data_for_display.dims[0], self.cursor[0]]]),
                                                   method='nearest')
        right_marginal_original = self.arr.sel(
            **dict([[self.data_for_display.dims[0], self.cursor[0]]]), method='nearest')
        figures['right_marginal'] = figure(plot_width=200, plot_height=self.app_main_size,
                                           title=None, y_range=figures['main'].y_range,
                                           x_range=(np.min(right_marginal.values), np.max(right_marginal.values)),
                                           y_axis_location='left', toolbar_location=None, tools=[])
        plots['right_marginal'] = figures['right_marginal'].line(
            y=right_marginal.coords[self.data_for_display.dims[1]].values, x=right_marginal.values)
        plots['right_marginal_original'] = figures['right_marginal'].line(
            y=right_marginal_original.coords[self.data_for_display.dims[1]].values, x=right_marginal_original.values,
            line_color='red')

        # add lines
        self.add_cursor_lines(figures['main'])
        _ = figures['main'].multi_line(xs=[], ys=[], line_color='white', line_width=1) # band lines

        # prep the widgets for the analysis function
        signature = inspect.signature(self.analysis_fn)

        # drop the first which has to be the input data, we can revisit later if this is too limiting
        parameter_names = list(signature.parameters)[1:]
        named_widgets = dict(zip(parameter_names, self.widget_specification))
        built_widgets = {}

        def update_marginals():
            right_marginal_data = self.data_for_display.sel(**dict([[self.data_for_display.dims[0], self.cursor[0]]]),
                                                            method='nearest')
            bottom_marginal_data = self.data_for_display.sel(**dict([[self.data_for_display.dims[1], self.cursor[1]]]),
                                                             method='nearest')
            plots['bottom_marginal'].data_source.data = {
                'x': bottom_marginal_data.coords[self.data_for_display.dims[0]].values,
                'y': bottom_marginal_data.values,
            }
            plots['right_marginal'].data_source.data = {
                'y': right_marginal_data.coords[self.data_for_display.dims[1]].values,
                'x': right_marginal_data.values,
            }

            right_marginal_data = self.arr.sel(**dict([[self.data_for_display.dims[0], self.cursor[0]]]),
                                                            method='nearest')
            bottom_marginal_data = self.arr.sel(**dict([[self.data_for_display.dims[1], self.cursor[1]]]),
                                                             method='nearest')
            plots['bottom_marginal_original'].data_source.data = {
                'x': bottom_marginal_data.coords[self.data_for_display.dims[0]].values,
                'y': bottom_marginal_data.values,
            }
            plots['right_marginal_original'].data_source.data = {
                'y': right_marginal_data.coords[self.data_for_display.dims[1]].values,
                'x': right_marginal_data.values,
            }
            figures['bottom_marginal'].y_range.start = np.min(bottom_marginal_data.values)
            figures['bottom_marginal'].y_range.end = np.max(bottom_marginal_data.values)
            figures['right_marginal'].x_range.start = np.min(right_marginal_data.values)
            figures['right_marginal'].x_range.end = np.max(right_marginal_data.values)

        def click_main_image(event):
            self.cursor = [event.x, event.y]
            update_marginals()

        error_msg = widgets.Div(text='')

        @Debounce(0.25)
        def update_data_for_display():
            try:
                self.data_for_display = self.analysis_fn(self.arr, *[built_widgets[p].value for p in parameter_names if p in built_widgets])
                error_msg.text = ''
            except Exception as e:
                error_msg.text = '{}'.format(e)

            # flush + update
            update_marginals()
            plots['main'].data_source.data = {
                'image': [self.data_for_display.values.T]
            }


        def update_data_change_wrapper(attr, old, new):
            if old != new:
                update_data_for_display()

        for parameter_name in named_widgets.keys():
            specification = named_widgets[parameter_name]

            widget = None
            if specification == int:
                widget = widgets.Slider(start=-20, end=20, value=0, title=parameter_name)
            if specification == float:
                widget = widgets.Slider(start=-20, end=20, value=0., step=0.1, title=parameter_name)

            if widget is not None:
                built_widgets[parameter_name] = widget
                widget.on_change('value', update_data_change_wrapper)

        update_main_colormap = self.update_colormap_for('main')

        self.app_context['run'] = lambda x: x

        main_color_range_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Range')

        # Attach callbacks
        main_color_range_slider.on_change('value', update_main_colormap)

        figures['main'].on_event(events.Tap, click_main_image)

        layout = row(column(figures['main'], figures['bottom_marginal']),
                     column(figures['right_marginal']),
                     column(
                         widgetbox(
                             *[built_widgets[p] for p in parameter_names if p in built_widgets]
                         ),
                         widgetbox(
                             self._cursor_info,
                             main_color_range_slider,
                             error_msg,
                         )
                     ))

        doc.add_root(layout)
        doc.title = 'Band Tool'


def dyn(dynamic_function: typing.Callable, data: DataType, widget_specifications=None):
    data = normalize_to_spectrum(data)

    tool = DynamicTool(dynamic_function, widget_specifications)
    return tool.make_tool(data)



import logging
import os
import re
import uuid
import warnings
from collections import namedtuple
from pathlib import Path
from typing import Iterator

import numpy as np
import pandas as pd

import arpes.config
from arpes.exceptions import ConfigurationError
from arpes.utilities.string import snake_case
from typing import Any, List, Optional, Union

__all__ = ['clean_xlsx_dataset', 'default_dataset', 'infer_data_path',
           'attach_extra_dataset_columns', 'swap_reference_map',
           'cleaned_dataset_exists', 'modern_clean_xlsx_dataset', 'cleaned_pair_paths',
           'list_files_for_rename', 'rename_files']

_DATASET_EXTENSIONS = {'.xlsx', '.xlx',}
_SEARCH_DIRECTORIES = ('', 'hdf5', 'fits', '../Data', '../Data/hdf5', '../Data/fits',)
_TOLERATED_EXTENSIONS = {'.h5', '.nc', '.fits', '.pxt', '.nxs', '.txt',}


def shorten_left(s, max_length=20):
    if len(s) > max_length:
        return '...' + s[-(max_length - 3):]


def list_files_for_rename(path=None, extensions=None):
    if path is None:
        path = os.getcwd()

    if extensions is None:
        extensions = ['.txt', '.nxs']

    files_list = list((Path(path).glob('*')))
    files_list = [p for p in files_list if p.suffix in extensions]
    return files_list


def rename_files(dry=True, path=None, extensions=None, starting_index=1):
    if path is None:
        path = os.getcwd()

    files = list_files_for_rename(path=path, extensions=extensions)
    for i, file in enumerate(files):
        extra = '_{}'.format(i + starting_index)
        new_file = file.parent / (file.stem + extra + file.suffix)

        short_input = shorten_left(str(file), max_length=20)
        short_output = shorten_left(str(file.parent / (file.stem + extra + file.suffix)),
                                    max_length=20 + len(extra))

        print('{} -> {}'.format(short_input, short_output))
        if not dry:
            os.rename(str(file), str(new_file))


def is_blank(item: Union[float, str]) -> bool:
    if isinstance(item, str):
        return item == ''

    if isinstance(item, float):
        return np.isnan(item)

    if pd.isnull(item):
        return True

    return False


def infer_data_path(file: int, scan_desc: pd.Series, allow_soft_match: bool = False, use_regex: bool = True) -> str:
    if not isinstance(file, str):
        file = str(int(file))

    if 'path' in scan_desc and not is_blank(scan_desc['path']):
        return scan_desc['path']

    workspace = arpes.config.CONFIG['WORKSPACE']
    workspace_path = os.path.join(workspace['path'], 'data')
    workspace = workspace['name']

    base_dir = workspace_path or os.path.join(arpes.config.DATA_PATH, workspace)
    dir_options = [os.path.join(base_dir, option) for option in _SEARCH_DIRECTORIES]

    # another plugin related option here is we can restrict the number of regexes by allowing plugins
    # to install regexes for particular endstations, if this is needed in the future it might be a good way
    # of preventing clashes where there is ambiguity in file naming scheme across endstations
    patterns = [
        r'[\-a-zA-Z0-9_\w+]+_[0]+{}_S[0-9][0-9][0-9]$'.format(file),
        r'[\-a-zA-Z0-9_\w+]+_{}_S[0-9][0-9][0-9]$'.format(file),
        r'[\-a-zA-Z0-9_\w+]+_[0]+{}_R[0-9][0-9][0-9]$'.format(file),
        r'[\-a-zA-Z0-9_\w+]+_{}_R[0-9][0-9][0-9]$'.format(file),
        r'[\-a-zA-Z0-9_\w+]+scan_[0]*{}_[0-9][0-9][0-9]'.format(file),
        r'[\-a-zA-Z0-9_\w+]+scan_[0]*{}'.format(file),
        r'[\-a-zA-Z0-9_\w]+_[0]+{}$'.format(file),
        r'[\-a-zA-Z0-9_\w]+_{}$'.format(file),
        r'[\-a-zA-Z0-9_\w]+{}$'.format(file),
        r'[\-a-zA-Z0-9_\w]+[0]{}$'.format(file),
        r'[\-a-zA-Z0-9_\w+]+_[0]+{}_S[0-9][0-9][0-9]$'.format(file),
        r'[\-a-zA-Z0-9_\w+]+_{}_S[0-9][0-9][0-9]$'.format(file)
    ]

    patterns = [re.compile(m) for m in patterns]

    for dir in dir_options:
        try:
            files = list(filter(lambda f: os.path.splitext(f)[1] in _TOLERATED_EXTENSIONS, os.listdir(dir)))
            if use_regex:
                for p in patterns:
                    for f in files:
                        m = p.match(os.path.splitext(f)[0])
                        if m is not None:
                            if m.string == os.path.splitext(f)[0]:
                                return os.path.join(dir, f)
            else:
                for f in files:
                    if os.path.splitext(file)[0] == os.path.splitext(f)[0]:
                        return os.path.join(dir, f)
                    if allow_soft_match:
                        matcher = os.path.splitext(f)[0].split('_')[-1]
                        try:
                            if int(matcher) == int(file):
                                return os.path.join(dir, f) # soft match
                        except ValueError:
                            pass
        except FileNotFoundError:
            pass

    if file and file[0] == 'f':  # try trimming the f off
        return infer_data_path(file[1:], scan_desc, allow_soft_match=allow_soft_match, use_regex=use_regex)

    raise ConfigurationError('Could not find file associated to {}'.format(file))


def swap_reference_map(df: pd.DataFrame, old_reference, new_reference):
    """
    Replaces instances of a reference map old_reference in the ref_map column with
    new_reference.
    :param df:
    :return:
    """
    df = df.copy()

    new_ref_id = df.loc[new_reference, ('id',)]

    for id, row in df.iterrows():
        if row.ref_map == old_reference:
            df.loc[id, ('ref_map', 'ref_id',)] = new_reference, new_ref_id

    return df


def default_dataset(workspace: Optional[Any] = None, match: Optional[str] = None, **kwargs: Any) -> pd.DataFrame:
    if workspace is not None:
        arpes.config.CONFIG['WORKSPACE'] = workspace

    workspace = arpes.config.CONFIG['WORKSPACE']

    if not workspace:
        raise ConfigurationError('You need to set the WORKSPACE attribute on CONFIG!')

    workspace_name = workspace['name']
    data_dir = os.path.join(arpes.config.DATASET_PATH, workspace_name)

    def is_dataset(filename: str) -> bool:
        if filename.startswith('~$') or filename.startswith('._'):
            # temporary files on Windows
            return False
        rest, ext = os.path.splitext(filename)
        rest, internal_ext = os.path.splitext(rest)

        return ext in _DATASET_EXTENSIONS and internal_ext != '.cleaned'

    candidates = list(filter(is_dataset, os.listdir(data_dir)))
    if match is not None:
        candidates = list(filter(lambda p: match in p, candidates))

    if (len(candidates)) > 1:
        print('Available candidates are:')
        print(candidates)

    assert len(candidates) == 1

    return clean_xlsx_dataset(os.path.join(data_dir, candidates[0]), **kwargs)


def attach_extra_dataset_columns(path, **kwargs):
    from arpes.io import load_dataset
    import arpes.xarray_extensions # pylint: disable=unused-import, redefined-outer-name

    base_filename, extension = os.path.splitext(path)
    if extension not in _DATASET_EXTENSIONS:
        logging.warning('File is not an excel file')
        return None

    if 'cleaned' in base_filename:
        new_filename = base_filename + extension
    else:
        new_filename = base_filename + '.cleaned' + extension
    assert os.path.exists(new_filename)

    ds = pd.read_excel(new_filename, **kwargs)

    ColumnDef = namedtuple('ColumnDef', ['default', 'source'])
    add_columns = {'spectrum_type': ColumnDef('', 'attr'), }

    for column, definition in add_columns.items():
        ds[column] = definition.default

    # Add required columns
    if 'id' not in ds:
        ds['id'] = np.nan

    if 'path' not in ds:
        ds['path'] = ''

    # Cascade blank values
    for index, row in ds.sort_index().iterrows():
        row = row.copy()

        print(row.id)
        try:
            scan = load_dataset(dataset_uuid=row.id, df=ds)
        except ValueError as e:
            logging.warning(str(e))
            logging.warning('Skipping {}! Unable to load scan.'.format(row.id))
            continue
        for column, definition in add_columns.items():
            if definition.source == 'accessor':
                ds.loc[index, (column,)] = getattr(scan.S, column)
            elif definition.source == 'attr':
                ds.loc[index, (column,)] = scan.attrs[column]

    os.remove(new_filename)
    excel_writer = pd.ExcelWriter(new_filename)
    ds.to_excel(excel_writer, index=False)
    excel_writer.save()

    return ds.set_index('file')


def with_inferred_columns(df: pd.DataFrame):
    """
    Attach inferred columns to a data frame representing an ARPES dataset.

    So far the columns attached are the reference map name, and the reference map_id
    :param df:
    :return: pd.DataFrame which is the union of columns in `df` and the columns produced here
    """

    df = df.copy()

    if 'spectrum_type' not in df.columns:
        warnings.warn('You need to have a spectrum_type column. Try using `prepare_raw_files` first.')
        return df

    df['ref_map'] = ''
    df['ref_id'] = ''

    last_map = None
    logging.warning('Assuming sort along index')
    for index, row in df.sort_index().iterrows():

        if last_map is not None:
            df.loc[index, ('ref_map', 'ref_id')] = last_map, df.loc[last_map, ('id',)]

        if row.spectrum_type == 'map':
            last_map = index

    return df


def cleaned_path(path: str) -> str:
    base_filename, extension = os.path.splitext(path)
    if '.cleaned' in base_filename:
        return base_filename + extension
    return base_filename + '.cleaned' + extension


def cleaned_pair_paths(path):
    base_filename, extension = os.path.splitext(path)
    if 'cleaned' in base_filename:
        return base_filename.replace('.cleaned', '') + extension, base_filename + extension

    return base_filename + extension, base_filename + '.cleaned' + extension


def cleaned_dataset_exists(path):
    return os.path.exists(cleaned_path(path))


def safe_read(path: str, **kwargs: Any) -> pd.DataFrame:
    REATTEMPT_LIMIT = 8
    skiprows = kwargs.pop('skiprows', None)

    def read_snake(x: Union[int, str]) -> Union[List[int], List[str]]:
        try:
            return [x, snake_case(x.strip())]
        except:
            try:
                return [x, x.strip()]
            except AttributeError:
                return [x, x]

    if skiprows is not None:
        read = pd.read_excel(path, skiprows=skiprows, **kwargs)

        return read.rename(columns=dict([read_snake(x) for x in list(read.columns)]))

    for skiprows in range(REATTEMPT_LIMIT):
        try:
            read = pd.read_excel(path, skiprows=skiprows, **kwargs)
            read = read.rename(columns=dict([read_snake(x) for x in list(read.columns)]))
            if 'file' in read.columns:
                return read
        except TypeError:
            # sometimes this happens due to what looks like a bug in pandas
            pass

    warnings.warn('You must supply both a `file` and a `location` column in your '
                  'spreadsheet in order to load data.')
    raise ValueError('Did you supply both a `file` and a `location` column in your spreadsheet? '
                     'Could not safely read dataset. Supply a `skiprows` parameter and check '
                     'the validity of your data.')


def modern_clean_xlsx_dataset(path, allow_soft_match=False, with_inferred_cols=True, write=False, **kwargs):
    original_path, cleaned_path_str = cleaned_pair_paths(path)
    original = safe_read(original_path, **kwargs)
    original = original[original.file.notnull()]
    cleaned = pd.DataFrame({'id': [], 'path': [], 'file': [], 'spectrum_type': []})
    if os.path.exists(cleaned_path_str):
        cleaned = safe_read(cleaned_path_str, skiprows=0, **kwargs)
        if 'file' in cleaned.columns:
            cleaned = cleaned[cleaned.file.notnull()]
        else:
            cleaned['file'] = cleaned.index

    joined = original.set_index('file').combine_first(cleaned.set_index('file'))

    last_index = None

    # Cascade blank values
    for index, row in joined.iterrows():
        row = row.copy()

        for key, value in row.iteritems():
            if key == 'id' and is_blank(row.id):
                joined.loc[index, ('id',)] = str(uuid.uuid1())

            elif key == 'path' and is_blank(value):
                joined.loc[index, ('path',)] = infer_data_path(index, row, allow_soft_match)

            elif last_index is not None and is_blank(value) and not is_blank(joined.loc[last_index, (key,)]):
                joined.loc[index, (key,)] = joined.loc[last_index, (key,)]

        last_index = index

    if write:
        excel_writer = pd.ExcelWriter(cleaned_path_str)
        joined.to_excel(excel_writer, index=False)
        excel_writer.save()

    if with_inferred_cols:
        return with_inferred_columns(joined)

    return joined


def clean_xlsx_dataset(path: str, allow_soft_match: bool = False, write: bool = True, with_inferred_cols: bool = True,
                       warn_on_exists: bool = False, **kwargs: Any) -> pd.DataFrame:
    reload = kwargs.pop('reload', False)
    _, extension = os.path.splitext(path)
    if extension not in _DATASET_EXTENSIONS:
        logging.warning('File is not an excel file')
        return None

    new_filename = cleaned_path(path)
    if os.path.exists(new_filename):
        if reload:
            if warn_on_exists:
                logging.warning('Cleaned dataset already exists! Removing...')

            os.remove(new_filename)
        else:
            if warn_on_exists:
                logging.warning('Cleaned dataset already exists! Reading existing...')
            ds = pd.read_excel(new_filename).set_index('file')
            if with_inferred_cols:
                return with_inferred_columns(ds)
            return ds

    ds = safe_read(path, **kwargs)
    ds.rename(columns=lambda c: c.lower().strip().replace(' ', '_'), inplace=True)

    column_renamings = {
        'photon_energy': 'hv',
        'temp': 'temperature',
        # 't': 'temperature', # This is ambiguous between temperature and delay!
    }
    ds.rename(columns=lambda c: column_renamings.get(c, c), inplace=True)

    ds = ds.loc[ds.index.dropna()]
    if 'path' not in ds.columns:
        ds = ds[pd.notnull(ds['file'])] # drop null files if path not specified
    else:
        try:
            ds = ds[pd.notnull(ds['file']) | pd.notnull('path')]
        except KeyError:
            pass

    last_index = None

    # Add required columns
    if 'id' not in ds:
        ds['id'] = np.nan

    if 'path' not in ds:
        ds['path'] = ''

    # Cascade blank values
    for index, row in ds.sort_index().iterrows():
        row = row.copy()

        for key, value in row.iteritems():
            if key == 'id' and is_blank(row.id):
                ds.loc[index, ('id',)] = str(uuid.uuid1())

            elif key == 'path' and is_blank(value):
                ds.loc[index, ('path',)] = infer_data_path(row['file'], row, allow_soft_match)

            elif last_index is not None and is_blank(value) and not is_blank(ds.loc[last_index, (key,)]):
                ds.loc[index, (key,)] = ds.loc[last_index, (key,)]

        last_index = index

    ds = ds.loc[:, ~ds.columns.str.contains('^unnamed:_')]

    if write:
        excel_writer = pd.ExcelWriter(new_filename)
        ds.to_excel(excel_writer, index=False)
        excel_writer.save()

    if with_inferred_cols:
        return with_inferred_columns(ds.set_index('file'))

    return ds.set_index('file')


def walk_datasets(skip_cleaned: bool = True, use_workspace: bool = False) -> Iterator[str]:
    root = os.getcwd()

    if use_workspace:
        root = arpes.config.CONFIG['WORKSPACE']['path']

    for path, _, files in os.walk(root):
        excel_files = [f for f in files if '.xlsx' in f or '.xlx' in f]

        for x in excel_files:
            if x[0] == '~':
                continue

            if skip_cleaned and 'cleaned' in os.path.join(path, x):
                continue

            print("├{}".format(x))
            yield os.path.join(path, x)



import math
import warnings

import matplotlib.pyplot as plt
import numpy as np

from arpes.fits import GStepBModel, broadcast_model
from arpes.provenance import save_plot_provenance
from arpes.utilities import apply_dataarray

from .utils import path_for_plot, label_for_dim

__all__ = ['fermi_edge_reference', 'plot_fit']

@save_plot_provenance
def plot_fit(data, title=None, axes=None, out=None, norm=None, **kwargs):
    """
    Plots the results of a fit of some lmfit model to some data.

    We introspect the model to determine which attributes we should plot,
    as well as their uncertainties
    :param data: The data, this should be of type DataArray<lmfit.model.ModelResult>
    :param title:
    :param ax:
    :param out:
    :param norm:
    :param kwargs:
    :return:
    """

    # get any of the elements
    reference_fit = data.values.ravel()[0]
    model = reference_fit.model
    SKIP_NAMES = {'const_bkg', 'lin_bkg', 'erf_amp'}
    param_names = [p for p in model.param_names if p not in SKIP_NAMES]
    n_params = len(param_names)
    MAX_COLS = 3
    n_rows = int(math.ceil(n_params / MAX_COLS))
    n_cols = n_params if n_params < MAX_COLS else MAX_COLS

    is_bootstrapped = 'bootstrap' in data.dims

    if axes is None:
        _, axes = plt.subplots(n_rows, n_cols, figsize=(15, 6))

    for i, param in enumerate(param_names):
        row = i // MAX_COLS
        column = i - (row * MAX_COLS)

        try:
            ax = axes[row, column]
        except IndexError:
            ax = axes[column] # n_rows = 1

        # extract the data for this param name
        # attributes are on .value and .stderr
        centers = data.T.map(lambda x: x.params[param].value)
        if is_bootstrapped:
            centers = centers.mean('bootstrap')

        centers.plot(ax=ax)

        ax.set_title('Fit var: {}'.format(param))

        if len(centers.dims) == 1:
            if is_bootstrapped:
                widths = data.T.map(lambda x: x.params[param].value).std('bootstrap')
            else:
                widths = data.T.map(lambda x: x.params[param].stderr)
            # then we can plot widths as well, otherwise we need more
            # figures, blergh
            x_coords = centers.coords[centers.dims[0]]
            ax.fill_between(x_coords, centers.values + widths.values,
                            centers.values - widths.values, alpha=0.5)


    if title is None:
        title = data.S.label.replace('_', ' ')

    # if multidimensional, we can share y axis as well
    #axes.set_xlabel(label_for_dim(data, axes[0][0].get_xlabel()))
    #ax.set_title(title, font_size=14)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()


@save_plot_provenance
def fermi_edge_reference(data, title=None, ax=None, out=None, norm=None, **kwargs):
    warnings.warn('Not automatically correcting for slit shape distortions to the Fermi edge')

    sum_dimensions = {'cycle', 'phi', 'kp', 'kx'}
    sum_dimensions.intersection_update(set(data.dims))
    summed_data = data.sum(*list(sum_dimensions))

    broadcast_dimensions = summed_data.dims
    broadcast_dimensions.remove('eV')
    if len(broadcast_dimensions) == 1:
        edge_fit = broadcast_model(GStepBModel, summed_data.sel(eV=slice(-0.1, 0.1)), broadcast_dimensions[0])
    else:
        warnings.warn('Could not product fermi edge reference. Too many dimensions: {}'.format(broadcast_dimensions))
        return

    centers = apply_dataarray(edge_fit, np.vectorize(lambda x: x.params['center'].value, otypes=[np.float]))
    widths = apply_dataarray(edge_fit, np.vectorize(lambda x: x.params['width'].value, otypes=[np.float]))

    if ax is None:
        _, ax = plt.subplots(figsize=(8, 5))

    if title is None:
        title = data.S.label.replace('_', ' ')

    plot_centers = centers.plot(norm=norm, ax=ax)
    plot_widths = widths.plot(norm=norm, ax=ax)

    ax.set_xlabel(label_for_dim(data, ax.get_xlabel()))
    ax.set_ylabel(label_for_dim(data, ax.get_ylabel()))

    ax.set_title(title, font_size=14)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()



import xarray as xr
from arpes.typing import DataType

__all__ = ('normalize_to_spectrum', 'normalize_to_dataset',)


def normalize_to_spectrum(data: DataType):
    from arpes.io import load_dataset
    if isinstance(data, xr.Dataset):
        if 'up' in data.data_vars:
            return data.up

        return data.S.spectrum

    if isinstance(data, str):
        return normalize_to_spectrum(load_dataset(dataset_uuid=data))

    # not guaranteed to be a spectrum, but close enough
    return data


def normalize_to_dataset(data: DataType):
    from arpes.io import load_dataset
    if isinstance(data, xr.Dataset):
        return data

    if isinstance(data, str):
        return load_dataset(dataset_uuid=data)



"""
For now we monkeypatch lmfit to make it easier to work with in Jupyter. We should consider forking
or providing a pull at a later date after this settles down.

The end goal here is to allow pleasing and functional representations of curve fitting sessions
performed in Jupyter, so that they can be rapidly understood, and screencapped for simple purposes,
like including in group meeting notes.
"""

# pylint: disable=protected-access

import numpy as np
from lmfit import model


def repr_html_ModelResult(self, **kwargs):
    """
    Provides a better Jupyter representation of an `lmfit.ModelResult` instance.
    :param self:
    :param kwargs:
    :return:
    """
    template = """
        <div>
            <span><strong>Converged: </strong>{success}</span>
            <div>{formatted_components}<div>
            <div>{parameters}</div>
        </div>
        """
    return template.format(
        success=self.success,
        formatted_components=''.join('<div>{}</div>'.format(c._repr_html_()) for c in self.components),
        parameters=self.params._repr_html_(**kwargs)
    )


def repr_html_Model(self):
    """
    Better Jupyter representation of `lmfit.Model` instances.
    :param self:
    :return:
    """
    template = """
    <div>
    <strong>{name}</strong>
    </div>
    """
    return template.format(name=self.name)


def repr_html_Parameters(self, short=False):
    skip_on_short = {'Min', 'Max', 'Vary', 'Expr', 'Brute_Step'}
    all = ['Name', 'Value', 'Min', 'Max', 'Stderr', 'Vary', 'Expr', 'Brute_Step']
    keys = list(self.keys())
    keys.sort()
    template = """
    <table>
      <thead>
          <tr>
            {cols}
          </tr>
      </thead>
      <tbody>
        {rows}
      </tbody>
    </table>
    """
    return template.format(
        cols=''.join('<th>{}</th>'.format(c) for c in all if not short or c not in skip_on_short),
        rows=''.join(self[p].to_table_row(short=short) for p in keys)
    )

def repr_html_Parameter(self, short=False):
    if short:
        return """
            <tr>
                <th>{name}</th>
                <th>{value:.3f}</th>
                <th>{stderr:.3f}</th>
            </tr>
            """.format(name=self.name, value=self.value, stderr=self.stderr,)

    template = """
            <tr>
                <th>{name}</th>
                <th>{value:.3f}</th>
                <th>{min:.3f}</th>
                <th>{max:.3f}</th>
                <th>{stderr:.3f}</th>
                <th>{vary}</th>
                <th>{expr}</th>
                <th>{brute_step}</th>
            </tr>
            """
    return template.format(
        name=self.name,
        value=self.value,
        min=self.min,
        max=self.max,
        stderr=self.stderr or np.inf,
        vary=self.vary,
        expr=self.expr or '',
        brute_step=self.brute_step or ''
    )


model.Model._repr_html_ = repr_html_Model
#model.Parameter._repr_html_ = repr_html_Parameter
model.Parameters._repr_html_ = repr_html_Parameters
model.ModelResult._repr_html_ = repr_html_ModelResult

model.Parameter.to_table_row = repr_html_Parameter

# we don't export anything, just monkey-patch
__all__ = tuple()



import operator
import warnings

import lmfit as lf
import numpy as np
from lmfit.models import guess_from_peak, update_param_vals
from scipy import stats
from scipy.ndimage import gaussian_filter
from scipy.signal import convolve
from scipy.special import erfc  # pylint: disable=no-name-in-module

import xarray as xr
from arpes.constants import HBAR_SQ_EV_PER_ELECTRON_MASS_ANGSTROM_SQ

__all__ = ('XModelMixin', 'FermiLorentzianModel', 'GStepBModel', 'QuadraticModel',
           'ExponentialDecayCModel', 'LorentzianModel', 'GaussianModel', 'VoigtModel',
           'ConstantModel', 'LinearModel', 'GStepBStandardModel', 'AffineBackgroundModel',
           'AffineBroadenedFD',
           'FermiDiracModel', 'BandEdgeBModel',
           'gaussian_convolve', 'TwoGaussianModel', 'TwoLorModel', 'TwoLorEdgeModel', 'SplitLorentzianModel')


class XModelMixin(lf.Model):
    """
    A mixin providing curve fitting for xarray.DataArray instances.

    This amounts mostly to making `lmfit` coordinate aware, and providing
    a translation layer between xarray and raw np.ndarray instances.

    Subclassing this mixin as well as an lmfit Model class should bootstrap
    an lmfit Model to one that works transparently on xarray data.

    Alternatively, you can use this as a model base in order to build new models.

    The core method here is `guess_fit` which is a convenient utility that performs both
    a `lmfit.Model.guess`, if available, before populating parameters and
    performing a curve fit.

    __add__ and __mul__ are also implemented, to ensure that the composite model
    remains an instance of a subclass of this mixin.
    """
    n_dims = 1
    dimension_order = None

    def guess_fit(self, data, params=None, weights=None, guess=True, debug=False, prefix_params=True, transpose=False,
                  **kwargs):
        """
        Params allows you to pass in hints as to what the values and bounds on parameters
        should be. Look at the lmfit docs to get hints about structure
        :param data:
        :param params:
        :param kwargs:
        :return:
        """
        if transpose:
            assert len(data.dims) == 1 and "You cannot transpose (invert) a multidimensional array (scalar field)."

        coord_values = {}
        if 'x' in kwargs:
            coord_values['x'] = kwargs.pop('x')

        real_data, flat_data = data, data

        new_dim_order = None
        if isinstance(data, xr.DataArray):
            real_data, flat_data = data.values, data.values
            assert len(real_data.shape) == self.n_dims

            if self.n_dims == 1:
                coord_values['x'] = data.coords[list(data.indexes)[0]].values
            else:
                def find_appropriate_dimension(dim_or_dim_list):
                    if isinstance(dim_or_dim_list, str):
                        assert dim_or_dim_list in data.dims
                        return dim_or_dim_list

                    else:
                        intersect = set(dim_or_dim_list).intersection(data.dims)
                        assert len(intersect) == 1
                        return list(intersect)[0]

                # resolve multidimensional parameters
                new_dim_order = [find_appropriate_dimension(dim_options) for dim_options in self.dimension_order]
                if list(new_dim_order) != list(data.dims):
                    warnings.warn('Transposing data for multidimensional fit.')
                    data = data.transpose(*new_dim_order)

                coord_values = {k: v.values for k, v in data.coords.items() if k in new_dim_order}
                real_data, flat_data = data.values, data.values.ravel()

        real_weights = weights
        if isinstance(weights, xr.DataArray):
            if self.n_dims == 1:
                real_weights = real_weights.values
            else:
                if new_dim_order is not None:
                    real_weights = weights.transpose(*new_dim_order).values.ravel()
                else:
                    real_weights = weights.values.ravel()

        if transpose:
            cached_coordinate = list(coord_values.values())[0]
            coord_values[list(coord_values.keys())[0]] = real_data
            real_data = cached_coordinate
            flat_data = real_data

        if guess:
            guessed_params = self.guess(real_data, **coord_values)
        else:
            guessed_params = self.make_params()

        if params is not None:
            for k, v in params.items():
                if isinstance(v, dict):
                    if prefix_params:
                        guessed_params[self.prefix + k].set(**v)
                    else:
                        guessed_params[k].set(**v)
            guessed_params.update({self.prefix + k: v for k, v in params.items() if isinstance(v, lf.model.Parameter)})

        result = None
        try:
            result = super().fit(flat_data, guessed_params, **coord_values, weights=real_weights, **kwargs)
            result.independent = coord_values
            result.independent_order = new_dim_order
        except Exception as e:
            print(e)
            if debug:
                import pdb
                pdb.post_mortem(e.__traceback__)
        finally:
            return result

    def xguess(self, data, **kwargs):
        x = kwargs.pop('x', None)

        real_data = data
        if isinstance(data, xr.DataArray):
            real_data = data.values
            assert len(real_data.shape) == 1
            x = data.coords[list(data.indexes)[0]].values

        return self.guess(real_data, x=x, **kwargs)

    def __add__(self, other):
        """+"""
        return XAdditiveCompositeModel(self, other, operator.add)

    def __mul__(self, other):
        """*"""
        return XMultiplicativeCompositeModel(self, other, operator.mul)


class XAdditiveCompositeModel(lf.CompositeModel, XModelMixin):
    """
    xarray coordinate aware composite model corresponding to the
    sum of two models.
    """

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()
        guessed = {}
        for c in self.components:
            guessed.update(c.guess(data, x=x, **kwargs))

        for k, v in guessed.items():
            pars[k] = v

        return pars


class XMultiplicativeCompositeModel(lf.CompositeModel, XModelMixin):
    """
    xarray coordinate aware composite model corresponding to the
    sum of two models.

    Currently this just copies +, might want to adjust things!
    """

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()
        guessed = {}
        for c in self.components:
            guessed.update(c.guess(data, x=x, **kwargs))

        for k, v in guessed.items():
            pars[k] = v

        return pars


class XConvolutionCompositeModel(lf.CompositeModel, XModelMixin):
    """
    Work in progress for convolving two `Model`s.
    """

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()
        guessed = {}

        for c in self.components:
            if c.prefix == 'conv_':
                # don't guess on the convolution term
                continue

            guessed.update(c.guess(data, x=x, **kwargs))

        for k, v in guessed.items():
            pars[k] = v

        return pars


def np_convolve(original_data, convolution_kernel):
    return convolve(original_data, convolution_kernel, mode='same')


def _convolve(original_data, convolution_kernel):
    n_points = min(len(original_data), len(convolution_kernel))
    padding = np.ones(n_points)
    temp = np.concatenate((padding * original_data[0], original_data, padding * original_data[-1]))
    convolved = np.convolve(temp, convolution_kernel, mode='valid')
    n_offset = int((len(convolved) - n_points) / 2)
    return (convolved[n_offset:])[:n_points]


def gaussian_convolve(model_instance):
    """
    Produces a model that consists of convolution with a Gaussian kernel
    :param model_instance:
    :return:
    """
    return XConvolutionCompositeModel(
        model_instance, GaussianModel(prefix='conv_'), convolve)


def effective_mass_bkg(eV, kp, m_star=0,
                       k_center=0, eV_center=0,
                       gamma=1, amplitude=1,
                       amplitude_k=0,
                       const_bkg=0, k_bkg=0, eV_bkg=0):
    """
    Model implementation function for simultaneous 2D curve fitting of band effective mass.
    Allows for an affine background in each dimension, together with variance in the band intensity
    along the band, as a very simple model of matrix elements. Together with prenormalizing your data
    this should allow reasonable fits of a lot of typical ARPES data.
    :param eV:
    :param k:
    :param m_star:
    :param k_center:
    :param eV_center:
    :param gamma:
    :param amplitude:
    :param amplitude_k:
    :param const_bkg:
    :param k_bkg:
    :param eV_bkg:
    :return:
    """
    bkg = np.outer(eV * 0 + 1, k_bkg * kp) + np.outer(eV_bkg * eV, kp * 0 + 1) + const_bkg

    # check units
    dk = kp - k_center
    offset = HBAR_SQ_EV_PER_ELECTRON_MASS_ANGSTROM_SQ * dk ** 2 / (2 * m_star + 1e-6)
    eVk = np.outer(eV, kp * 0 + 1)
    coherent = (amplitude + amplitude_k * dk) * (1 / (2 * np.pi)) * gamma / (
        (eVk - eV_center + offset) ** 2 + (0.5 * gamma) ** 2)
    return (coherent + bkg).ravel()


def affine_bkg(x, lin_bkg=0, const_bkg=0):
    """
    An affine/linear background.
    :param x:
    :param lin_bkg:
    :param const_bkg:
    :return:
    """
    return lin_bkg * x + const_bkg


def quadratic(x, a=1, b=0, c=0):
    return a * x ** 2 + b * x + c


def gstepb(x, center=0, width=1, erf_amp=1, lin_bkg=0, const_bkg=0):
    """
    Fermi function convoled with a Gaussian together with affine background.
    This accurately represents low temperature steps where thermal broadening is
    less substantial than instrumental resolution.

    :param x: value to evaluate function at
    :param center: center of the step
    :param width: width of the step
    :param erf_amp: height of the step
    :param lin_bkg: linear background slope
    :param const_bkg: constant background
    :return:
    """
    dx = x - center
    return const_bkg + lin_bkg * np.min(dx, 0) + gstep(x, center, width, erf_amp)


def gstep(x, center=0, width=1, erf_amp=1):
    """
    Fermi function convolved with a Gaussian
    :param x: value to evaluate fit at
    :param center: center of the step
    :param width: width of the step
    :param erf_amp: height of the step
    :return:
    """
    dx = x - center
    return erf_amp * 0.5 * erfc(1.66511 * dx / width)


def gstepb_standard(x, center=0, sigma=1, amplitude=1, **kwargs):
    return gstepb(x, center, width=sigma, erf_amp=amplitude, **kwargs)


def exponential_decay_c(x, amp, tau, t0, const_bkg):
    """
    Represents an exponential decay after a point (delta) impulse.
    This coarsely models the dynamics after excitation in a
    pump-probe experiment.
    :param x:
    :param amp:
    :param tau:
    :param t0:
    :param const_bkg:
    :return:
    """
    dx = x - t0
    mask = (dx >= 0) * 1
    return const_bkg + amp * mask * np.exp(-(x - t0) / tau)


def twoexponential_decay_c(x, amp, t0, tau1, tau2, const_bkg):
    """
    Like `exponential_decay_c`, except that two timescales corresponding
    for instance to different quasiparticle decay channels are allowed,
    represented by `tau1` and `tau2`.
    :param x:
    :param amp:
    :param t0:
    :param tau1:
    :param tau2:
    :param const_bkg:
    :return:
    """
    dx = x - t0
    mask = (dx >= 0) * 1
    y = const_bkg + amp * (1 - np.exp(-dx / tau1)) * np.exp(-dx / tau2)
    f = y.copy()
    f[dx < 0] = const_bkg
    f[dx >= 0] = y[dx >= 0]
    return f


def lorentzian(x, gamma, center, amplitude):
    return amplitude * (1 / (2 * np.pi)) * gamma / ((x - center) ** 2 + (.5 * gamma) ** 2)


def pseudo_shirley(x, gamma, center, amplitude, bkg_amplitude):
    """
    WIP for a lorentzian with a "shirley like" background.
    :param x:
    :param gamma:
    :param center:
    :param amplitude:
    :param bkg_amplitude:
    :return:
    """
    lor = lorentzian(x, gamma, center, amplitude)


def twolorentzian(x, gamma, t_gamma, center, t_center, amp, t_amp, lin_bkg, const_bkg):
    """
    A double lorentzian model. This is typically not necessary, as you can use the
    + operator on the Model instances. For instance `LorentzianModel() + LorentzianModel(prefix='b')`.

    This mostly exists for people that prefer to do things the "Igor Way".
    :param x:
    :param gamma:
    :param t_gamma:
    :param center:
    :param t_center:
    :param amp:
    :param t_amp:
    :param lin_bkg:
    :param const_bkg:
    :return:
    """
    L1 = lorentzian(x, gamma, center, amp)
    L2 = lorentzian(x, t_gamma, t_center, t_amp)
    AB = affine_bkg(x, lin_bkg, const_bkg)
    return L1 + L2 + AB


def gstepb_mult_lorentzian(x, center=0, width=1, erf_amp=1, lin_bkg=0, const_bkg=0, gamma=1, lorcenter=0):
    return gstepb(x, center, width, erf_amp, lin_bkg, const_bkg) * lorentzian(x, gamma, lorcenter, 1)


def fermi_dirac(x, center=0, width=0.05, scale=1):
    """
    Fermi edge
    :param x:
    :param center:
    :param width:
    :param scale:
    :return:
    """
    return scale / (np.exp((x - center) / width) + 1)


def fermi_dirac_bkg(x, center=0, width=0.05, lin_bkg=0, const_bkg=0, scale=1):
    # Fermi edge with an affine background multiplied in
    return (scale + lin_bkg) / (np.exp((x - center) / width) + 1) + const_bkg


def band_edge_bkg(x, center=0, width=0.05, amplitude=1, gamma=0.1, lor_center=0, offset=0, lin_bkg=0, const_bkg=0):
    # Lorentzian plus affine background multiplied into fermi edge with overall offset
    return (lorentzian(x, gamma, lor_center, amplitude) + lin_bkg * x + const_bkg) * fermi_dirac(x, center,
                                                                                                 width) + offset


def lorentzian_affine(x, gamma=1, lor_center=0, amplitude=1, lin_bkg=0, const_bkg=0):
    return (lorentzian(x, gamma, lor_center, amplitude) + lin_bkg * x + const_bkg)


def gaussian(x, center=0, sigma=1, amplitude=1):
    return amplitude * np.exp(-(x - center) ** 2 / (2 * sigma ** 2))


def twogaussian(x, center=0, t_center=0, width=1, t_width=1, amp=1, t_amp=1, lin_bkg=0, const_bkg=0):
    return gaussian(x, center, width, amp) + gaussian(x, t_center, t_width, t_amp) + affine_bkg(x, lin_bkg, const_bkg)


def twolorentzian_gstep(x, gamma, t_gamma, center, t_center, amp, t_amp, lin_bkg, const_bkg, g_center, sigma, erf_amp):
    TL = twolorentzian(x, gamma, t_gamma, center, t_center, amp, t_amp, lin_bkg, const_bkg)
    GS = gstep(x, g_center, sigma, erf_amp)
    return TL * GS


def affine_broadened_fd(x, fd_center=0, fd_width=0.003, conv_width=0.02, const_bkg=1, lin_bkg=0, offset=0):
    """
    Fermi function convoled with a Gaussian together with affine background
    :param x: value to evaluate function at
    :param center: center of the step
    :param width: width of the step
    :param erf_amp: height of the step
    :param lin_bkg: linear background slope
    :param const_bkg: constant background
    :return:
    """
    dx = x - fd_center
    x_scaling = x[1] - x[0]
    fermi = 1 / (np.exp(dx / fd_width) + 1)
    return gaussian_filter(
        (const_bkg + lin_bkg * dx) * fermi,
        sigma=conv_width / x_scaling
    ) + offset


def log_renormalization(x, kF=1.6, kD=1.6, kC=1.7, alpha=0.4, vF=1e6):
    """
    Logarithmic correction to linear dispersion of materials with Dirac dispersion near charge neutrality.

    As examples, this can be used to study the low energy physics in high quality ARPES spectra of graphene
    or topological Dirac semimetals.
    :param k: value to evaluate fit at
    :param kF: Fermi wavevector
    :param kD: Dirac point
    :param alpha: Fine structure constant
    :param vF: Bare Band Fermi Velocity
    :param kC: Cutoff Momentum
    """
    dk = x - kF
    dkD = x - kD
    return -vF * np.abs(dkD) + (alpha / 4) * vF * dk * np.log(np.abs(kC / dkD))


def dirac_dispersion(x, kd=1.6, amplitude_1=1, amplitude_2=1, center=0, sigma_1=1, sigma_2=1):
    """
    Model for dirac_dispersion symmetric about the dirac point. Fits lorentziants to (kd-center) and (kd+center)
    :param x: value to evaluate fit at
    :param kd: Dirac point momentum
    :param amplitude_1: amplitude of Lorentzian at kd-center
    :param amplitude_2: amplitude of Lorentzian at kd+center
    :param center: center of Lorentzian
    :param sigma_1: FWHM of Lorentzian at kd-center
    :param sigma_2: FWHM of Lorentzian at kd+center
    """
    return lorentzian(x, center=kd - center, amplitude=amplitude_1, gamma=sigma_1) + \
           lorentzian(x, center=kd + center, amplitude=amplitude_2, gamma=sigma_2)


def g(x, mu=0, sigma=0.1):
    return (1 / np.sqrt(2 * np.pi * sigma ** 2)) * np.exp(-(1 / 2) * ((x - mu) / sigma) ** 2)


def band_edge_bkg_gauss(x, center=0, width=0.05, amplitude=1, gamma=0.1, lor_center=0, offset=0, lin_bkg=0,
                        const_bkg=0):  # ,sigma=0.1):
    return np_convolve(band_edge_bkg(x, 0, width, amplitude, gamma, lor_center, offset, lin_bkg, const_bkg),
                       g(np.linspace(-6, 6, 800), 0, 0.01))


def fermi_dirac_affine(x, center=0, width=0.05, lin_bkg=0, const_bkg=0, scale=1):
    # Fermi edge with an affine background multiplied in
    return (scale + lin_bkg * x) / (np.exp((x - center) / width) + 1) + const_bkg


def fermi_dirac_bkg_gauss(x, center=0, width=0.05, lin_bkg=0, const_bkg=0, scale=1, sigma=0.01):
    return np_convolve(
        fermi_dirac_affine(x, center, width, lin_bkg, const_bkg, scale),
        # g(np.linspace(-6,6,800),0,sigma)
        g(x, (min(x) + max(x)) / 2, sigma)
    )


def gstep_stdev(x, center=0, sigma=1, erf_amp=1):
    """
    Fermi function convolved with a Gaussian
    :param x: value to evaluate fit at
    :param center: center of the step
    :param width: width of the step
    :param erf_amp: height of the step
    :return:
    """
    dx = x - center
    return erf_amp * 0.5 * erfc(np.sqrt(2) * dx / sigma)


def gstepb_stdev(x, center=0, sigma=1, erf_amp=1, lin_bkg=0, const_bkg=0):
    """
    Fermi function convoled with a Gaussian together with affine background
    :param x: value to evaluate function at
    :param center: center of the step
    :param width: width of the step
    :param erf_amp: height of the step
    :param lin_bkg: linear background slope
    :param const_bkg: constant background
    :return:
    """
    dx = x - center
    return const_bkg + lin_bkg * np.min(dx, 0) + gstep_stdev(x, center, sigma, erf_amp)


class EffectiveMassModel(XModelMixin):
    """
    A two dimensional model for a quadratic distribution of Lorentzians.
    This can be used for "global" fitting of the effective mass of a band, providing higher quality
    results than iterative fitting for the lineshapes and then performing a
    quadratic fit to the centers.

    This model also provides a representative example of how to implement (2+)D "global"
    lmfit.Model classes which are xarray coordinate aware.

        `dimension_order`: allows specifying allowed sets of dimension for the global fit (here eV
        must be one axis, while either kp or phi is allowed for the other.
    """

    n_dims = 2
    dimension_order = ['eV', ['kp', 'phi']]

    def __init__(self, independent_vars=('eV', 'kp',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(effective_mass_bkg, **kwargs)

        self.set_param_hint('gamma', min=0.)
        self.set_param_hint('amplitude', min=0.)

    def guess(self, data, eV=None, kp=None, phi=None, **kwargs):
        momentum = kp if kp is not None else phi
        try:
            momentum = momentum.values
            eV = eV.values
            data = data.values
        except AttributeError:
            pass

        pars = self.make_params()

        pars['%sm_star' % self.prefix].set(value=1)

        pars['%sk_center' % self.prefix].set(value=np.mean(momentum))
        pars['%seV_center' % self.prefix].set(value=np.mean(eV))

        pars['%samplitude' % self.prefix].set(value=np.mean(np.mean(data, axis=0)))
        pars['%sgamma' % self.prefix].set(value=0.25)

        pars['%samplitude_k' % self.prefix].set(value=0)  # can definitely improve here

        # Crude estimate of the background
        left, right = np.mean(data[:5, :], axis=0), np.mean(data[-5:, :], axis=0)
        top, bottom = np.mean(data[:, :5], axis=0), np.mean(data[:, -5:], axis=0)
        left, right = np.percentile(left, 10), np.percentile(right, 10)
        top, bottom = np.percentile(top, 10), np.percentile(bottom, 10)

        pars['%sconst_bkg' % self.prefix].set(value=np.min(np.array([left, right, top, bottom])))
        pars['%sk_bkg' % self.prefix].set(value=(bottom - top) / (eV[-1] - eV[0]))
        pars['%seV_bkg' % self.prefix].set(value=(right - left) / (momentum[-1] - momentum[0]))

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class AffineBroadenedFD(XModelMixin):
    """
    A model for fitting an affine density of states with resolution broadened Fermi-Dirac occupation
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(affine_broadened_fd, **kwargs)

        self.set_param_hint('offset', min=0.)
        self.set_param_hint('fd_width', min=0.)
        self.set_param_hint('conv_width', min=0.)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%sfd_center' % self.prefix].set(value=0)
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.mean().item() * 2)
        pars['%soffset' % self.prefix].set(value=data.min().item())

        pars['%sfd_width' % self.prefix].set(0.005)  # TODO we can do better than this
        pars['%sconv_width' % self.prefix].set(0.02)

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class FermiLorentzianModel(XModelMixin):
    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(gstepb_mult_lorentzian, **kwargs)

        self.set_param_hint('erf_amp', min=0.)
        self.set_param_hint('width', min=0)
        self.set_param_hint('lin_bkg', min=-10, max=10)
        self.set_param_hint('const_bkg', min=-50, max=50)
        self.set_param_hint('gamma', min=0.)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%scenter' % self.prefix].set(value=0)
        pars['%slorcenter' % self.prefix].set(value=0)
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%swidth' % self.prefix].set(0.02)  # TODO we can do better than this
        pars['%serf_amp' % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class FermiDiracModel(XModelMixin):
    """
    A model for the Fermi Dirac function
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='drop', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(fermi_dirac, **kwargs)

        self.set_param_hint('width', min=0)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['{}center'.format(self.prefix)].set(value=0)
        pars['{}width'.format(self.prefix)].set(value=0.05)
        pars['{}scale'.format(self.prefix)].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class GStepBModel(XModelMixin):
    """
    A model for fitting Fermi functions with a linear background
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(gstepb, **kwargs)

        self.set_param_hint('erf_amp', min=0.)
        self.set_param_hint('width', min=0)
        self.set_param_hint('lin_bkg', min=-10, max=10)
        self.set_param_hint('const_bkg', min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%scenter' % self.prefix].set(value=0)
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%swidth' % self.prefix].set(0.02)  # TODO we can do better than this
        pars['%serf_amp' % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class TwoBandEdgeBModel(XModelMixin):
    """
    A model for fitting a Lorentzian and background multiplied into the fermi dirac distribution
    TODO, actually implement two_band_edge_bkg (find original author and their intent)
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({
            'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars,
        })
        super().__init__(two_band_edge_bkg, **kwargs)

        self.set_param_hint('amplitude_1', min=0.)
        self.set_param_hint('gamma_1', min=0.)
        self.set_param_hint('amplitude_2', min=0.)
        self.set_param_hint('gamma_2', min=0.)

        self.set_param_hint('offset', min=-10)

    def guess(self, data, x=None, **kwargs):
        # should really do some peak fitting or edge detection to find
        # okay values here
        pars = self.make_params()

        if x is not None:
            slope, intercept, r_value, p_value, std_err = stats.linregress(x, data)
            pars['%slor_center' % self.prefix].set(value=x[np.argmax(data - slope * x)])
        else:
            pars['%slor_center' % self.prefix].set(value=-0.2)

        pars['%sgamma' % self.prefix].set(value=0.2)
        pars['%samplitude' % self.prefix].set(value=(data.mean() - data.min()) / 1.5)

        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%soffset' % self.prefix].set(value=data.min())

        pars['%scenter' % self.prefix].set(value=0)
        pars['%swidth' % self.prefix].set(0.02)  # TODO we can do better than this

        return update_param_vals(pars, self.prefix, **kwargs)


class BandEdgeBModel(XModelMixin):
    """
    A model for fitting a Lorentzian and background multiplied into the fermi dirac distribution
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({
            'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars,
        })
        super().__init__(band_edge_bkg, **kwargs)

        self.set_param_hint('amplitude', min=0.)
        self.set_param_hint('gamma', min=0.)
        self.set_param_hint('offset', min=-10)

    def guess(self, data, x=None, **kwargs):
        # should really do some peak fitting or edge detection to find
        # okay values here
        pars = self.make_params()

        if x is not None:
            slope, intercept, r_value, p_value, std_err = stats.linregress(x, data)
            pars['%slor_center' % self.prefix].set(value=x[np.argmax(data - slope * x)])
        else:
            pars['%slor_center' % self.prefix].set(value=-0.2)

        pars['%sgamma' % self.prefix].set(value=0.2)
        pars['%samplitude' % self.prefix].set(value=(data.mean() - data.min()) / 1.5)

        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%soffset' % self.prefix].set(value=data.min())

        pars['%scenter' % self.prefix].set(value=0)
        pars['%swidth' % self.prefix].set(0.02)  # TODO we can do better than this

        return update_param_vals(pars, self.prefix, **kwargs)


class BandEdgeBGModel(XModelMixin):
    """
    A model for fitting a Lorentzian and background multiplied into the fermi dirac distribution
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({
            'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars,
        })
        super().__init__(band_edge_bkg_gauss, **kwargs)

        self.set_param_hint('amplitude', min=0.)
        self.set_param_hint('gamma', min=0.)
        self.set_param_hint('offset', min=-10)
        self.set_param_hint('center', vary=False)

    def guess(self, data, x=None, **kwargs):
        # should really do some peak fitting or edge detection to find
        # okay values here
        pars = self.make_params()

        if x is not None:
            slope, intercept, r_value, p_value, std_err = stats.linregress(x, data)
            pars['%slor_center' % self.prefix].set(value=x[np.argmax(data - slope * x)])
        else:
            pars['%slor_center' % self.prefix].set(value=-0.2)

        pars['%sgamma' % self.prefix].set(value=0.2)
        pars['%samplitude' % self.prefix].set(value=(data.mean() - data.min()) / 1.5)

        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%soffset' % self.prefix].set(value=data.min())

        # pars['%scenter' % self.prefix].set(value=0)
        pars['%swidth' % self.prefix].set(0.02)  # TODO we can do better than this

        return update_param_vals(pars, self.prefix, **kwargs)


class FermiDiracAffGaussModel(XModelMixin):
    """
    A model for the Fermi Dirac function with an affine background multiplied, then all convolved with a Gaussian
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='drop', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(fermi_dirac_bkg_gauss, **kwargs)

        # self.set_param_hint('width', min=0)
        self.set_param_hint('width', vary=False)
        # self.set_param_hint('lin_bkg', max=10)
        # self.set_param_hint('scale', max=50000)
        self.set_param_hint('scale', min=0)
        self.set_param_hint('sigma', min=0, vary=True)
        self.set_param_hint('lin_bkg', vary=False)
        self.set_param_hint('const_bkg', vary=False)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['{}center'.format(self.prefix)].set(value=0)
        # pars['{}width'.format(self.prefix)].set(value=0.05)
        pars['{}width'.format(self.prefix)].set(value=0.0009264)
        pars['{}scale'.format(self.prefix)].set(value=data.mean() - data.min())
        pars['{}lin_bkg'.format(self.prefix)].set(value=0)
        pars['{}const_bkg'.format(self.prefix)].set(value=0)
        pars['{}sigma'.format(self.prefix)].set(value=0.023)

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class GStepBStdevModel(XModelMixin):
    """
    A model for fitting Fermi functions with a linear background
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(gstepb_stdev, **kwargs)

        self.set_param_hint('erf_amp', min=0.)
        self.set_param_hint('sigma', min=0)
        self.set_param_hint('lin_bkg', min=-10, max=10)
        self.set_param_hint('const_bkg', min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%scenter' % self.prefix].set(value=0)
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%ssigma' % self.prefix].set(0.02)  # TODO we can do better than this
        pars['%serf_amp' % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class GStepBStandardModel(XModelMixin):
    """
    A model for fitting Fermi functions with a linear background
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(gstepb_standard, **kwargs)

        self.set_param_hint('amplitude', min=0.)
        self.set_param_hint('sigma', min=0)
        self.set_param_hint('lin_bkg', min=-10, max=10)
        self.set_param_hint('const_bkg', min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%scenter' % self.prefix].set(value=0)
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%ssigma' % self.prefix].set(0.02)  # TODO we can do better than this
        pars['%samplitude' % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class ExponentialDecayCModel(XModelMixin):
    """
    A model for fitting an exponential decay with a constant background
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(exponential_decay_c, **kwargs)

        # amp is also a parameter, but we have no hint for it
        self.set_param_hint('tau', min=0.)
        # t0 is also a parameter, but we have no hint for it
        self.set_param_hint('const_bkg')

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%stau' % self.prefix].set(value=0.2)  # 200fs
        pars['%st0' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.mean())
        pars['%samp' % self.prefix].set(value=data.max() - data.mean())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class TwoExponentialDecayCModel(XModelMixin):
    """
    A model for fitting an exponential decay with a constant background
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(twoexponential_decay_c, **kwargs)

        # amp is also a parameter, but we have no hint for it
        self.set_param_hint('tau1', min=0.)
        self.set_param_hint('tau2', min=0.)
        # t0 is also a parameter, but we have no hint for it
        self.set_param_hint('const_bkg')

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%stau1' % self.prefix].set(value=0.2)  # 200fs
        pars['%stau2' % self.prefix].set(value=1)  # 1ps
        pars['%st0' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.mean())
        pars['%samp' % self.prefix].set(value=data.max() - data.mean())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class QuadraticModel(XModelMixin):
    """
    A model for fitting a quadratic function
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(quadratic, **kwargs)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%sa' % self.prefix].set(value=0)
        pars['%sb' % self.prefix].set(value=0)
        pars['%sc' % self.prefix].set(value=data.mean())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class AffineBackgroundModel(XModelMixin):
    """
    A model for an affine background
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(affine_bkg, **kwargs)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%slin_bkg' % self.prefix].set(value=np.percentile(data, 10))
        pars['%sconst_bkg' % self.prefix].set(value=0)

        return update_param_vals(pars, self.prefix, **kwargs)


class TwoGaussianModel(XModelMixin):
    """
    A model for two gaussian functions with a linear background
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(twogaussian, **kwargs)

        self.set_param_hint('amp', min=0.)
        self.set_param_hint('width', min=0)
        self.set_param_hint('t_amp', min=0.)
        self.set_param_hint('t_width', min=0)
        self.set_param_hint('lin_bkg', min=-10, max=10)
        self.set_param_hint('const_bkg', min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%scenter' % self.prefix].set(value=0)
        pars['%st_center' % self.prefix].set(value=0)
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%swidth' % self.prefix].set(0.02)  # TODO we can do better than this
        pars['%st_width' % self.prefix].set(0.02)
        pars['%samp' % self.prefix].set(value=data.mean() - data.min())
        pars['%st_amp' % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class TwoLorModel(XModelMixin):
    """
    A model for two gaussian functions with a linear background
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(twolorentzian, **kwargs)

        self.set_param_hint('amp', min=0.)
        self.set_param_hint('gamma', min=0)
        self.set_param_hint('t_amp', min=0.)
        self.set_param_hint('t_gamma', min=0)
        self.set_param_hint('lin_bkg', min=-10, max=10)
        self.set_param_hint('const_bkg', min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%scenter' % self.prefix].set(value=0)
        pars['%st_center' % self.prefix].set(value=0)
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%sgamma' % self.prefix].set(0.02)  # TODO we can do better than this
        pars['%st_gamma' % self.prefix].set(0.02)
        pars['%samp' % self.prefix].set(value=data.mean() - data.min())
        pars['%st_amp' % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class TwoLorEdgeModel(XModelMixin):
    """
    A model for (two lorentzians with an affine background) multiplied by a gstepb
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(twolorentzian_gstep, **kwargs)

        self.set_param_hint('amp', min=0.)
        self.set_param_hint('gamma', min=0)
        self.set_param_hint('t_amp', min=0.)
        self.set_param_hint('t_gamma', min=0)
        self.set_param_hint('erf_amp', min=0.)
        self.set_param_hint('sigma', min=0)
        self.set_param_hint('lin_bkg', min=-10, max=10)
        self.set_param_hint('const_bkg', min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%scenter' % self.prefix].set(value=0)
        pars['%st_center' % self.prefix].set(value=0)
        pars['%sg_center' % self.prefix].set(value=0)
        pars['%slin_bkg' % self.prefix].set(value=0)
        pars['%sconst_bkg' % self.prefix].set(value=data.min())
        pars['%sgamma' % self.prefix].set(0.02)  # TODO we can do better than this
        pars['%st_gamma' % self.prefix].set(0.02)
        pars['%ssigma' % self.prefix].set(0.02)
        pars['%samp' % self.prefix].set(value=data.mean() - data.min())
        pars['%st_amp' % self.prefix].set(value=data.mean() - data.min())
        pars['%serf_amp' % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class Log_Renormalization_Model(XModelMixin):
    """
    A model for Logarithmic Renormalization to Linear Dispersion in Dirac Materials
    :param k: value to evaluate fit at
    :param kF: Fermi wavevector
    :param kD: Dirac point
    :param alpha: Fine structure constant
    :param vF: Bare Band Fermi Velocity
    :param kC: Cutoff Momentum
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(log_renormalization, **kwargs)

        self.set_param_hint('alpha', min=0.)
        self.set_param_hint('vF', min=0.)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        pars['%skC' % self.prefix].set(value=1.7)
        # pars['%svF' % self.prefix].set(value=(data.max()-data.min())/(kC-kD))

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class DiracDispersionModel(XModelMixin):
    """
    Model for dirac_dispersion symmetric about the dirac point. Fits lorentziants to (kd-center) and (kd+center)
    :param x: value to evaluate fit at
    :param kd: Dirac point momentum
    :param amplitude_1: amplitude of Lorentzian at kd-center
    :param amplitude_2: amplitude of Lorentzian at kd+center
    :param center: center of Lorentzian
    :param sigma_1: FWHM of Lorentzian at kd-center
    :param sigma_2: FWHM of Lorentzian at kd+center
    """

    def __init__(self, independent_vars=('x',), prefix='', missing='raise', name=None, **kwargs):
        kwargs.update({'prefix': prefix, 'missing': missing, 'independent_vars': independent_vars})
        super().__init__(dirac_dispersion, **kwargs)

        self.set_param_hint('sigma_1', min=0.)
        self.set_param_hint('sigma_2', min=0.)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()

        # pars['%skd' % self.prefix].set(value=1.5)


        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class LorentzianModel(XModelMixin, lf.models.LorentzianModel):
    pass


class SplitLorentzianModel(XModelMixin, lf.models.SplitLorentzianModel):
    def _set_paramhints_prefix(self):
        """
        In lmfit v0.9.13 there is a bug here where the prefix is not set
        on the parameter hint expressions for the computed variables "height"
        and "fwhm". Until this is patched, we need to do this correctly, which invovles
        just injecting the prefix in before each of the model parameters in these
        expressions.
        """
        self.set_param_hint('sigma', min=0)
        self.set_param_hint('sigma_r', min=0)
        self.set_param_hint('fwhm', expr='{prefix}sigma+{prefix}sigma_r'.format(
            prefix=self.prefix))
        self.set_param_hint(
            'height', expr='2*{prefix}amplitude/{:.7f}/max({}, ({prefix}sigma+{prefix}sigma_r))'.format(
                np.pi, np.finfo(np.float).eps, prefix=self.prefix))

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()
        """Estimate initial model parameter values from data."""
        pars = guess_from_peak(self, data, x, negative=False, ampscale=1.25)
        sigma = pars['%ssigma' % self.prefix]
        pars['%ssigma_r' % self.prefix].set(value=sigma.value, min=sigma.min, max=sigma.max)

        return update_param_vals(pars, self.prefix, **kwargs)


class VoigtModel(XModelMixin, lf.models.VoigtModel):
    pass


class GaussianModel(XModelMixin, lf.models.GaussianModel):
    pass


class ConstantModel(XModelMixin, lf.models.ConstantModel):
    pass


class LinearModel(XModelMixin, lf.models.LinearModel):
    def guess(self, data, x=None, **kwargs):
        sval, oval = 0., 0.
        if x is not None:
            sval, oval = np.polyfit(x, data, 1)
        pars = self.make_params(intercept=oval, slope=sval)
        return update_param_vals(pars, self.prefix, **kwargs)



import numpy as np
import scipy.ndimage.filters

import arpes.fits
import xarray as xr
from arpes.analysis.band_analysis_utils import (param_getter,
                                                param_stderr_getter)

__all__ = ['Band', 'MultifitBand', 'VoigtBand', 'BackgroundBand', 'FermiEdgeBand',
           'AffineBackgroundBand']


class Band:
    def __init__(self, label, display_label=None, data=None):
        self.label = label
        self._display_label = display_label
        self._data = data

    @property
    def display_label(self):
        return self._display_label or self.label

    @display_label.setter
    def display_label(self, value):
        self._display_label = value

    @property
    def velocity(self):
        spacing = float(self.coords[self.dims[0]][1] - self.coords[self.dims[0]][0])

        def embed_nan(values, padding):
            embedded = np.ndarray((values.shape[0] + 2 * padding,))
            embedded[:] = float('nan')
            embedded[padding:-padding] = values
            return embedded

        raw_values = embed_nan(np.copy(self.center.values), 50)


        masked = np.copy(raw_values)
        masked[raw_values != raw_values] = 0

        nan_mask = np.copy(raw_values) * 0 + 1
        nan_mask[raw_values != raw_values] = 0

        sigma = 0.1 / spacing
        nan_mask = scipy.ndimage.gaussian_filter(nan_mask, sigma, mode='mirror')
        masked = scipy.ndimage.gaussian_filter(masked, sigma, mode='mirror')

        return xr.DataArray(
            np.gradient(masked / nan_mask, spacing)[50:-50],
            self.coords,
            self.dims
        )

    @property
    def fermi_velocity(self):
        return self.velocity.sel(eV=0, method='nearest')

    @property
    def band_width(self):
        return None

    def band_energy(self, coordinates):
        pass

    @property
    def self_energy(self):
        return None

    @property
    def fit_cls(self):
        return arpes.fits.LorentzianModel

    def get_dataarray(self, var_name, clean=True):
        if not clean:
            return self._data[var_name].values

        output = np.copy(self._data[var_name].values)
        output[self._data[var_name + '_stderr'].values > 0.01] = float('nan')

        return xr.DataArray(
            output,
            self._data[var_name].coords,
            self._data[var_name].dims,
        )

    @property
    def center(self, clean=True):
        return self.get_dataarray('center', clean)

    @property
    def center_stderr(self, clean=False):
        return self.get_dataarray('center_stderr', clean)

    @property
    def sigma(self, clean=True):
        return self.get_dataarray('sigma', clean)

    @property
    def amplitude(self, clean=True):
        return self.get_dataarray('amplitude', clean)

    @property
    def indexes(self):
        return self._data.center.indexes

    @property
    def coords(self):
        return self._data.center.coords

    @property
    def dims(self):
        return self._data.center.dims


class MultifitBand(Band):
    """
    Convenience class that reimplements reading data out of a composite fit result
    """

    def get_dataarray(self, var_name, clean=True):
        full_var_name = self.label + var_name

        if 'stderr' in full_var_name:
            return self._data.T.map(param_stderr_getter(full_var_name.split('_stderr')[0]))

        return self._data.T.map(param_getter(full_var_name))


class VoigtBand(Band):
    @property
    def fit_cls(self):
        return arpes.fits.VoigtModel


class BackgroundBand(Band):
    @property
    def fit_cls(self):
        return arpes.fits.GaussianModel


class FermiEdgeBand(Band):
    @property
    def fit_cls(self):
        return arpes.fits.GStepBStandardModel


class AffineBackgroundBand(Band):
    @property
    def fit_cls(self):
        return arpes.fits.AffineBackgroundModel



import numpy as np
import scipy.ndimage.interpolation

import colorcet as cc
import xarray as xr
from arpes.plotting.interactive_utils import BokehInteractiveTool
from arpes.typing import DataType
from arpes.utilities.funcutils import Debounce
from arpes.utilities.normalize import normalize_to_spectrum

__all__ = ['ComparisonTool', 'compare']


class ComparisonTool(BokehInteractiveTool):
    auto_zero_nans = False
    auto_rebin = False
    other = None
    compared = None

    def __init__(self, other, **kwargs):
        super().__init__()

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get('app_main_size', 600)
        self.other = other

    def tool_handler(self, doc):
        from bokeh.layouts import row, column, widgetbox
        from bokeh.models import widgets
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.plotting import figure

        default_palette = self.default_palette
        difference_palette = cc.coolwarm

        intensity_slider = widgets.Slider(
            title='Relative Intensity Scaling', start=0.5, end=1.5,
            step=0.005, value=1)

        self.app_context.update({
            'A': self.arr,
            'B': self.other,
            'compared': self.compared,
            'plots': {},
            'figures': {},
            'widgets': {},
            'data_range': self.arr.T.range(),
            'color_maps': {},
        })

        self.color_maps['main'] = LinearColorMapper(
            default_palette, low=np.min(self.arr.values), high=np.max(self.arr.values), nan_color='black')

        figure_kwargs = {
            'tools': ['reset', 'wheel_zoom'],
            'plot_width': self.app_main_size,
            'plot_height': self.app_main_size,
            'min_border': 10,
            'toolbar_location': 'left',
            'x_range': self.data_range['x'],
            'y_range': self.data_range['y'],
            'x_axis_location': 'below',
            'y_axis_location': 'right',
        }

        self.figures['A'] = figure(title='Spectrum A', **figure_kwargs)
        self.figures['B'] = figure(title='Spectrum B', **figure_kwargs)
        self.figures['compared'] = figure(title='Comparison', **figure_kwargs)

        self.compared = self.arr - self.other
        diff_low, diff_high = np.min(self.arr.values), np.max(self.arr.values)
        diff_range = np.sqrt((abs(diff_low) + 1) * (abs(diff_high) + 1)) * 1.5
        self.color_maps['difference'] = LinearColorMapper(
            difference_palette, low=-diff_range, high=diff_range, nan_color='white')

        self.plots['A'] = self.figures['A'].image(
            [self.arr.values], x=self.data_range['x'][0], y=self.data_range['y'][0],
            dw=self.data_range['x'][1] - self.data_range['x'][0],
            dh=self.data_range['y'][1] - self.data_range['y'][0],
            color_mapper=self.color_maps['main'],
        )

        self.plots['B'] = self.figures['B'].image(
            [self.other.values], x=self.data_range['x'][0], y=self.data_range['y'][0],
            dw=self.data_range['x'][1] - self.data_range['x'][0],
            dh=self.data_range['y'][1] - self.data_range['y'][0],
            color_mapper=self.color_maps['main']
        )

        self.plots['compared'] = self.figures['compared'].image(
            [self.compared.values], x=self.data_range['x'][0], y=self.data_range['y'][0],
            dw=self.data_range['x'][1] - self.data_range['x'][0],
            dh=self.data_range['y'][1] - self.data_range['y'][0],
            color_mapper=self.color_maps['difference']
        )

        x_axis_name = self.arr.dims[0]
        y_axis_name = self.arr.dims[1]

        stride = self.arr.T.stride()
        delta_x_axis = stride['x']
        delta_y_axis = stride['y']

        delta_x_slider = widgets.Slider(
            title='{} Shift'.format(x_axis_name), start=-20 * delta_x_axis,
            step=delta_x_axis / 2, end=20 * delta_x_axis, value=0)

        delta_y_slider = widgets.Slider(
            title='{} Shift'.format(y_axis_name), start=-20 * delta_y_axis,
            step=delta_y_axis / 2, end=20 * delta_y_axis, value=0)

        @Debounce(0.5)
        def update_summed_figure(attr, old, new):
            # we don't actually use the args because we need to pull all the data out
            shifted = (intensity_slider.value) * scipy.ndimage.interpolation.shift(self.other.values, [
                delta_x_slider.value / delta_x_axis,
                delta_y_slider.value / delta_y_axis,
            ], order=1, prefilter=False, cval=np.nan)
            self.compared = self.arr - xr.DataArray(
                shifted,
                coords=self.arr.coords,
                dims=self.arr.dims)

            self.compared.attrs.update(**self.arr.attrs)
            try:
                del self.compared.attrs['id']
            except KeyError:
                pass

            self.app_context['compared'] = self.compared
            self.plots['compared'].data_source.data = {
                'image': [self.compared.values]
            }

        layout = column(
            row(
                column(self.app_context['figures']['A']),
                column(self.app_context['figures']['B']),
            ),
            row(
                column(self.app_context['figures']['compared']),
                widgetbox(
                    intensity_slider,
                    delta_x_slider,
                    delta_y_slider,
                ),
            )
        )

        update_summed_figure(None, None, None)

        delta_x_slider.on_change('value', update_summed_figure)
        delta_y_slider.on_change('value', update_summed_figure)
        intensity_slider.on_change('value', update_summed_figure)

        doc.add_root(layout)
        doc.title = 'Comparison Tool'


def compare(A: DataType, B: DataType):
    A = normalize_to_spectrum(A)
    attrs = A.attrs
    B = normalize_to_spectrum(B)

    # normalize total intensity
    TOTAL_INTENSITY = 1000000
    A = A / (A.sum(A.dims) / TOTAL_INTENSITY)
    B = B / (B.sum(B.dims) / TOTAL_INTENSITY)
    A.attrs.update(**attrs)

    tool = ComparisonTool(other=B)
    return tool.make_tool(A)
